from extractor import extractor as ext

import os
import numpy
import zfit
import math
import pprint
import pytest
import utils_noroot as utnr
import rk.utilities as rkut

from logzero    import logger    as log
from cmb_ck     import combiner  as cmb_ck
from np_reader  import np_reader as np_rdr
from mc_reader  import mc_reader as mc_rdr
from cs_reader  import cs_reader as cs_rdr
from rk_model   import rk_model  as model
from cmb_ck     import combiner  as cmb_ck

#-----------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return
#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#----------------------------------------------------
def check_close(value, target, abs_tol=1e-4):
    pas_val = math.isclose(value, target, abs_tol = abs_tol)
    if not pas_val:
        log.error(f'{value:.6f} != {target:.6f}')
        raise
#----------------------------------------------------
def test_simple():
    log.info('Running: test_simple')
    d_yld       = {'all' : (2e3, 2e2) } 
    obs_mm_sp   = zfit.Space('mass mm', limits=(2600, 3300))

    mod         = model(preffix='simple', d_nent=d_yld, obs_mm_sp=obs_mm_sp)
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()
    d_val, d_var= mod.get_cons()

    obj         = ext()
    obj.ck      = 0.1 
    obj.cov     = 0.01 ** 2
    obj.data    = d_dat
    obj.model   = d_mod 
    obj.const   = d_val, d_var
    obj.plt_dir = 'tests/extractor/simple'

    result      = obj.get_fit_result()
    result.hesse()

    print(result)

    delete_all_pars()
#----------------------------------------------------
def test_fix_pars():
    skip_test()
    log.info('Running: test_const')
    l_dset = ['all_TOS']

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache_dir= './v65_v63_v24' 
    rdr.cache    = True
    cv_sys       = rdr.get_cov(kind='sys')
    cv_sta       = rdr.get_cov(kind='sta')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod          = model(preffix='const', d_eff=d_eff, d_nent=d_rare_yld, l_dset=l_dset)
    d_mod        = mod.get_model()
    d_dat        = mod.get_data()
    d_val, d_var = mod.get_cons()

    if l_dset == ['all_TOS'] or l_dset == ['all_TOS', 'all_TIS']:
        cmb                 = cmb_ck(rk=1, eff=d_eff, yld=d_rare_yld)
        cmb.out_dir         = 'plots/combination'
        t_comb              = cmb.get_combination()
        d_rjpsi, d_eff, cov = t_comb
    else:
        cov = cv_sys + cv_sta

    cov = numpy.array([[cov[0][0]]])

    obj          = ext(dset=l_dset)
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.cov      = cov 
    obj.fix      = ['dmu_ee', 'dmu_mm', 'ssg_ee', 'ssg_mm', 'r0_ee', 'r1_ee', 'r2_ee', 'ncb_mm', 'ncb_ee', 'npr_ee'] 
    obj.const    = d_val, d_var
    obj.plt_dir  = 'tests/extractor/fix_pars'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    print(result)

    delete_all_pars()
#----------------------------------------------------
def test_json():
    log.info('Running: test_json')
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e3, 'd2' :            2e3, 'd3' :            2e3, 'd4' :            2e3} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = model(preffix='simple', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_yld)
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()
    d_pre       = mod.get_prefit_pars()

    obj         = ext(dset=['2017_TOS', '2018_TOS'])
    obj.eff     = d_eff 
    obj.data    = d_dat
    obj.model   = d_mod 
    obj.plt_dir = 'tests/extractor/json'

    result      = obj.get_fit_result()
    result.hesse()
    d_pos = rkut.result_to_dict(result)

    utnr.dump_json({'pre' : d_pre, 'pos' : d_pos}, 'tests/extractor/json/pars.json')

    delete_all_pars()
#----------------------------------------------------
def test_efficiency():
    log.info('Running: test_efficiency')
    d_eff        = model.get_eff(kind='diff')
    d_yld        = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 
    d_mcmu       = {'d1' :   (50, 49), 'd2' :   (51, 49), 'd3' :   (51, 48), 'd4' :   (52, 51)}
    d_mcsg       = {'d1' :    (2,  4), 'd2' :   (1, 1.8), 'd3' :     (2, 3), 'd4' :     (3, 4)} 

    mod          = model(preffix='efficiency', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg)
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext()
    obj.eff      = d_eff 
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/efficiency'

    result       = obj.get_fit_result()
    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']

    delete_all_pars()
#----------------------------------------------------
def test_constraint():
    log.info('Running: test_constraint')
    d_eff        = model.get_eff(kind='half')
    d_yld        = {'d1' :          2e3, 'd2' :            2e3, 'd3' :            2e3, 'd4' :            2e3} 
    d_mcmu       = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg       = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff        = rename_keys(d_eff)
    d_yld        = rename_keys(d_yld, use_txs=False)
    d_mcmu       = rename_keys(d_mcmu)
    d_mcsg       = rename_keys(d_mcsg)

    mod          = model(preffix='constraint', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_yld)
    cvmat        = mod.get_cov(kind='random')
    d_dat        = mod.get_data()
    d_mod        = mod.get_model()

    obj          = ext(dset=['2017_TOS', '2018_TOS'])
    obj.eff      = d_eff 
    obj.cov      = cvmat
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/constraint'

    result       = obj.get_fit_result()
    result.hesse()
    print(result)

    delete_all_pars()
#----------------------------------------------------
def test_rjpsi():
    log.info('Running: test_rjpsi')
    d_eff        = model.get_eff(kind='half')
    d_yld        = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 
    d_mcmu       = {'d1' :   (50, 49), 'd2' :   (51, 49), 'd3' :   (51, 48), 'd4' :   (52, 51)}
    d_mcsg       = {'d1' :    (2,  4), 'd2' :   (1, 1.8), 'd3' :     (2, 3), 'd4' :     (3, 4)} 

    mod          = model(preffix='rjpsi', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg)
    cvmat        = mod.get_cov(kind='diag_eq', c=0.001)
    d_rjpsi      = mod.get_rjpsi(kind='eff_bias')
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext()
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cvmat
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/rjpsi'

    result       = obj.get_fit_result()

    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']

    delete_all_pars()
#----------------------------------------------------
def test_real():
    skip_test()
    log.info('Running: test_real')

    rdr          = np_rdr(sys='v65', sta='v81', yld='v24')
    d_rare_yld   = rdr.get_ryields()

    mod          = model(preffix='real', d_nent=d_rare_yld)
    mod.out_dir  = 'tests/extractor/real/model'
    d_mod        = mod.get_model()
    d_dat        = mod.get_data()
    d_val, d_var = mod.get_cons()

    obj          = ext()
    obj.cov      = 0.01 ** 2 
    obj.data     = d_dat
    obj.const    = d_val, d_var
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/real'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    utnr.dump_pickle(result, 'tests/extractor/real/result.pkl')

    delete_all_pars()
#----------------------------------------------------
def test_all_years():
    skip_test()
    log.info('Running: test_real')

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    cv_sys       = rdr.get_cov(kind='sys')
    cv_sta       = rdr.get_cov(kind='sta')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod          = model(preffix='real', d_eff=d_eff, d_nent=d_rare_yld, l_dset=['all_TOS'])
    mod.out_dir  = 'tests/extractor/all_years/model'
    d_mod        = mod.get_model()
    d_dat        = mod.get_data()
    d_val, d_var = mod.get_cons()

    obj          = ext()
    #--------------------------
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cv_sys + cv_sta
    #--------------------------
    obj.data     = d_dat
    obj.const    = d_val, d_var
    obj.model    = d_mod 
    #--------------------------
    obj.plt_dir  = 'tests/extractor/all_years'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    utnr.dump_pickle(result, 'tests/extractor/all_years/result.pkl')

    delete_all_pars()
#----------------------------------------------------
def test_combined():
    skip_test()
    log.info('Running: test_real')

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    cv_sys       = rdr.get_cov(kind='sys')
    cv_sta       = rdr.get_cov(kind='sta')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod          = model(preffix='real', d_eff=d_eff, d_nent=d_rare_yld, l_dset=['all_TOS', 'all_TIS'])
    mod.out_dir  = 'tests/extractor/combined/model'
    d_mod        = mod.get_model()
    d_dat        = mod.get_data()
    d_val, d_var = mod.get_cons()

    cmb                 = cmb_ck(rjp=d_rjpsi, eff=d_eff, cov=cv_sys + cv_sta)
    cmb.out_dir         = 'tests/extractor/combined/combination'
    d_rjpsi, d_eff, cov = cmb.get_combination(add_tis=True)

    obj          = ext(dset=['all_TOS', 'all_TIS'])
    #--------------------------
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cov 
    #--------------------------
    obj.model    = d_mod 
    obj.data     = d_dat
    obj.const    = d_val, d_var
    #--------------------------
    obj.plt_dir  = 'tests/extractor/combined'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    utnr.dump_pickle(result, 'tests/extractor/combined/result.pkl')

    delete_all_pars()
#----------------------------------------------------
def test_real_const():
    skip_test()
    log.info('Running: test_const')

    rdr          = cs_rdr(version='v4', preffix='const')
    d_val, d_var = rdr.get_constraints()

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    cv_sys       = rdr.get_cov(kind='sys')
    cv_sta       = rdr.get_cov(kind='sta')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    rdr          = mc_rdr(version='v4')
    rdr.cache    = True 
    d_mcmu       = rdr.get_parameter(name='mu')
    d_mcsg       = rdr.get_parameter(name='sg')

    mod                  = model(preffix='const', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_rare_yld)
    d_mod                = mod.get_model()
    d_dat                = mod.get_data()
    d_val_mod, d_var_mod = mod.get_cons()

    d_val.update(d_val_mod)
    d_var.update(d_var_mod)

    obj          = ext(dset=['r2p1_TOS', '2017_TOS', '2018_TOS'])
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cv_sys + cv_sta
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.const    = d_val, d_var
    obj.plt_dir  = 'tests/extractor/const'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    print(result)

    delete_all_pars()
#----------------------------------------------------
def test_real_const_dset():
    log.info('Running: test_dset')

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod          = model(preffix='dset', d_eff=d_eff, d_nent=d_rare_yld, l_dset=['all_TOS'])
    mod.bdt_bin  = 5
    mod.kind     = 'nom'
    d_val, d_var = mod.get_cons()
    d_mod        = mod.get_model()
    d_dat        = mod.get_data()

    obj          = ext(dset=['all_TOS'])
    obj.cov      = numpy.array([[0.23]])
    obj.const    = d_val, d_var
    obj.model    = d_mod 
    obj.eff      = d_eff
    obj.data     = d_dat
    obj.rjpsi    = d_rjpsi
    obj.plt_dir  = 'tests/extractor/dset'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    utnr.dump_pickle(result, 'tests/extractor/dset/result.pkl')

    print(result)
    delete_all_pars()
#----------------------------------------------------
def test_diagonal():
    log.info('Running: test_constraint')
    d_eff        = model.get_eff(kind='half')
    d_yld        = {'d1' :          2e3, 'd2' :            2e3, 'd3' :            2e3, 'd4' :            2e3} 
    d_mcmu       = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg       = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff        = rename_keys(d_eff)
    d_yld        = rename_keys(d_yld, use_txs=False)
    d_mcmu       = rename_keys(d_mcmu)
    d_mcsg       = rename_keys(d_mcsg)

    mod          = model(preffix='constraint', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg)
    cvmat        = mod.get_cov(kind='random')
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext(dset=['2017_TOS', '2018_TOS'], drop_correlations=True)
    obj.eff      = d_eff 
    obj.cov      = cvmat
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/constraint'

    result       = obj.get_fit_result()

    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']

    delete_all_pars()
#----------------------------------------------------
def test_dmodel():
    log.info('Running: test_simple')
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e2, 'd2' :            2e2, 'd3' :            2e2, 'd4' :            2e2} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = model(preffix='simple', d_eff=d_eff, d_nent=d_yld, l_dset=['2018_TOS'])
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    obj         = ext(dset=['2018_TOS'])
    obj.eff     = d_eff 
    obj.data    = d_dat
    obj.model   = d_mod 
    obj.plt_dir = 'tests/extractor/simple'

    result      = obj.get_fit_result()
    result.hesse()

    print(result)

    delete_all_pars()
#----------------------------------------------------
def main():
    utnr.timer_on=True
    test_simple()
    return
    test_real_const_dset()
    test_real()
    test_fix_pars()
    test_dmodel()
    test_combined()
    test_all_years()
    test_real_const()
    test_json()
    test_constraint()
    test_diagonal()
    test_rjpsi()
    test_efficiency()
#----------------------------------------------------
if __name__ == '__main__':
    main()




from scales    import scales          as scl
from scales    import get_proc_labels 
from log_store import log_store

import os
import mplhep
import pandas            as pnd
import utils_noroot      as utnr
import matplotlib.pyplot as plt


log=log_store.add_logger('rk_extractor:test_scales')
#-------------------------------
def plot_df(df, trig):
    df = df[df.trig == trig]
    ax = None
    d_proc_lab = get_proc_labels()
    for proc, df_p in df.groupby('kind'):
        ax=df_p.plot(x='year', y='val', yerr='err', ax=ax, label=d_proc_lab[proc], linestyle='none', marker='o')

    os.makedirs('tests/scales/all_datasets', exist_ok=True)

    plt_path = f'tests/scales/all_datasets/{trig}.png'
    log.info(f'Saving to: {plt_path}')
    plt.grid()
    plt.ylabel('Scale')
    plt.ylim(0.0, 0.5)
    plt.xlabel('')
    plt.tight_layout()
    plt.savefig(plt_path)
    plt.close('all')
#-------------------------------
def test_all_datasets():
    df = pnd.DataFrame(columns=['year', 'trig', 'kind', 'val', 'err'])
    for year in ['2011', '2012', '2015', '2016', '2017', '2018']:
        #for trig in ['ETOS', 'GTIS']:
        for trig in ['ETOS']:
            for kind in ['bpks', 'bdks', 'bsph']:#, 'bpk1', 'bpk2']:
                obj      = scl(dset=year, trig=trig, kind=kind)
                val, err = obj.get_scale()

                df = utnr.add_row_to_df(df, [year, trig, kind, val, err])

    plot_df(df, 'ETOS')
    #plot_df(df, 'GTIS')
#-------------------------------
def test_simple():
    obj      = scl(dset='2011', trig='ETOS', kind='bpks')
    val, err = obj.get_scale()
#-------------------------------
def main():
    plt.style.use(mplhep.style.LHCb2)
    test_simple()
    test_all_datasets()
#-------------------------------
if __name__ == '__main__':
    main()




import numpy
import zfit
import math
import re
import os
import pprint
import zutils.utils      as zut
import matplotlib.pyplot as plt
import utils_noroot      as utnr

from scipy.stats           import poisson
from bdt_scale             import scale_reader as scl_rdr
from log_store             import log_store

log = log_store.add_logger(name='rk_extractor:rkex_model')
#----------------------------------------------------
class model:
    def __init__(self, rk=1, preffix='', d_nent=None, channel=None, obs_mm_sp=None):
        '''
        Parameters
        -----------------------
        rk     (float): Value used in model, will impact toy data produced 
        preffix(str)  : Used to name parameters in case multiple models used
        d_nent (dict) : Stores number of B -> llK entries, after selection, e.g. {r1 : 23235}
        l_dset (list) : List of datasets for which the models should be created, if None, do it for all datasets i.e. r1_TOS, r2p1_TOS...
        channel (str) : If used, PDFs and datasets will be restricted to this channel, values 'ee', 'mm'
        obs_mm_sp (zfit.space): Zfit observable to be used in the double swapped di-muon mass.
        '''
        self._rk         = rk
        self._preffix    = preffix
        self._d_nent     = d_nent
        self._chan       = channel 
        self._obs_mm_sp  = obs_mm_sp

        self._l_all_kind = None

        zfit.settings.changed_warnings.hesse_name = False

        self._nbin        = 80
        self._msid_vers   = 'v4'
        self._comb_vers   = 'v1'
        self._d_lumi      = {'r1' : 3, 'r2p1' : 1.9, '2017' : 1.7, '2018' : 2.1}
        self._l_dset      = ['r1', 'r2p1', '2017', '2018']
        self._d_mod       = None
        self._out_dir     = None
        self._mod_dir     = None
        self._kind        = 'nom' 
        self._d_rare_scl  = {}

        self._initialized = False
    #----------------------------------------------------
    def _initialize(self):
        if self._initialized:
            return
        log.info('Starting initialization')

        if self._d_nent is None:
            log.error(f'No B-> Jpsi K yields found')
            raise

        self._set_kinds()
        self._check_kind()
        self._cache_model()
        self._validate_model()

        log.info('Finished initializing')

        if self._out_dir is not None:
            utnr.dump_json(self._d_rare_scl, f'{self._out_dir}/rare_scales.json')

        self._initialized = True
    #----------------------------------------------------
    def _validate_model(self):
        for ds, (mod_mm, mod_ee) in self._d_mod.items(): 
            self._plot_model(f'{ds}_mm', mod_mm)
            self._plot_model(f'{ds}_ee', mod_ee)

            self._print_model(f'{ds}_mm', mod_mm)
            self._print_model(f'{ds}_ee', mod_ee)
    #---------------------------------------------------------------
    def _set_kinds(self):
        log.debug(f'Setting model kinds')
        self._l_all_kind = []
        self._l_all_kind.append(r'sig_MTOS:sys\d{1}')
        self._l_all_kind.append(r'cmb_MTOS:sys\d{1}')

        self._l_all_kind.append(r'sig_ETOS:sys\d{1}')
        self._l_all_kind.append(r'cpr_ETOS:sys\d{1}')
        self._l_all_kind.append(r'cmb_ETOS:sys\d{1}')

        self._l_all_kind.append(r'rpr_ETOS:bts\d+')
        self._l_all_kind.append(r'cpr_ETOS:bts\d+')
    #---------------------------------------------------------------
    def _check_kind(self):
        log.debug(f'Checking model kinds')
        if   self._kind is None:
            log.error(f'Kind not specified')
            raise
        elif self._kind == 'nom':
            log.debug('Using nominal model')
            return

        found = False
        for kind in self._l_all_kind:
            try:
                if re.match(kind, self._kind):
                    found = True
                    break
            except:
                log.error(f'Cannot match {self._kind} to {kind}')
                raise

        if not found: 
            log.error(f'Model of kind {self._kind} not supported')
            raise
        else:
            log.debug(f'Using model of kind: {self._kind}')
    #----------------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create: {value}')
            raise

        self._out_dir = value
        self._mod_dir = f'{self._out_dir}/plots/models'
        os.makedirs(self._mod_dir, exist_ok=True)

        log.debug(f'Using output directory: {self._out_dir}')
    #----------------------------------------------------
    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, value):
        '''
        Parameters
        -----------------
        kind (str): Will define the type of PDF, e.g. "signal:sys1", if not specified, will use nominal model
        '''
        self._kind = value
    #----------------------------------------------------
    def _get_ds_model(self, ds, nent_mm, nent_ee):
        if self._chan is not None:
            log.warning(f'Picking up model for channel: {self._chan}')
        else:
            log.info('Getting model for both channels')

        if   self._chan is None:
            pdf_mm = self._get_pdf(preffix=f'mm_{ds}', nent=nent_mm)
            pdf_ee = self._get_pdf(preffix=f'ee_{ds}', nent=nent_ee)
        elif self._chan == 'mm':
            pdf_mm = self._get_pdf(preffix=f'mm_{ds}', nent=nent_mm)
            pdf_ee = None
        elif self._chan == 'ee':
            pdf_mm = None
            pdf_ee = self._get_pdf(preffix=f'ee_{ds}', nent=nent_ee)
        else:
            log.error(f'Invalid channel: {self._chan}')
            raise

        return pdf_mm, pdf_ee
    #----------------------------------------------------
    def _cache_model(self):
        log.info('Caching models')
        d_nent = { dset : (nent_mm * self._rk, nent_ee) for dset, [nent_mm, nent_ee] in self._d_nent.items() }
        d_mod  = {}

        for dset, (nent_mm, nent_ee) in d_nent.items():
            mod_mm, mod_ee = self._get_ds_model(dset, nent_mm, nent_ee)
            d_mod[dset]    = mod_mm, mod_ee

        self._d_mod = d_mod
    #----------------------------------------------------
    def get_model(self):
        '''
        Returns
        -----------------
        d_ent (dict): Returns a dictionary: {name : tup} where
            name (str)  : model identifier, e.g. r1
            tup  (tuple): Tuple with muon and electron PDFs, e.g. pdf_mm, pdf_ee 
        '''

        self._initialize()

        return self._d_mod
    #----------------------------------------------------
    def _add_ext_constraints(self, d_par, d_var):
        if d_var is None:
            log.warning(f'Not adding errors for constrained parameters in prefit dictionary')
            return d_par

        d_par_new = {}
        for name, var in d_var.items():
            if name not in d_par:
                log.error(f'Cannot find {name} in prefit dictionary:')
                pprint.pprint(d_par.keys())
                raise

            val = d_par[name][0]
            err = math.sqrt(var)

            d_par_new[name] = [val, err]

        d_par.update(d_par_new)

        return d_par
    #----------------------------------------------------
    def _add_ck(self, d_par):
        regex='nsg_ee_([0-9a-z]+_.*)'
        d_par_ck = {}
        for var_name in d_par:
            mtch = re.match(regex, var_name)
            if not mtch:
                continue

            suffix = mtch.group(1)

            ee_yld_name = var_name
            mm_yld_name = var_name.replace('_ee_', '_mm_')

            ee_yld, _   = d_par[ee_yld_name]
            mm_yld, _   = d_par[mm_yld_name]

            d_par_ck[f'ck_{suffix}'] = [ (self._rk * ee_yld) / mm_yld, 0]

        d_par.update(d_par_ck)

        return d_par
    #----------------------------------------------------
    def _add_ck_constraints(self, d_par, ck_var):
        if ck_var is None:
            log.warning(f'Not adding errors for ck parameters in prefit dictionary')
            return d_par

        try:
            [ck_name] = [ name for name in d_par if name.startswith('ck_') ] 
        except:
            log.error('Expected one and only one CK parameter in:')
            pprint.pprint(d_par)
            raise

        [ck_val, _] = d_par[ck_name]
        ck_err      = math.sqrt(ck_var)

        d_par[ck_name] = [ck_val, ck_err]

        return d_par
    #----------------------------------------------------
    def get_prefit_pars(self, d_var=None, ck_var=None):
        '''
        Used to get model parameters used to make the toy data

        Parameters
        --------------------
        d_var (dict): Dictionary with variances for parameters that are constrained. If pased the
        constraint widths will be added as errors in the prefit dictionary

        ck_var (float): CK variance after combining different datasets 

        Returns 
        --------------------
        d_par (dict): Dictionary storing the prefit parameters (used to build the model) and their
        errors, e.g. {'par_x' : (3, 1)}
        '''
        self._initialize()

        d_model = self.get_model()

        d_par = {}
        for l_mod_mm, mod_ee in d_model.values():
            if l_mod_mm is not None:
                [ mod_bm_mm, mod_qm_mm ] = l_mod_mm
                d_par_bm_mm = { par.name : [ par.value().numpy(), 0] for par in mod_bm_mm.get_params() }
                d_par_qm_mm = { par.name : [ par.value().numpy(), 0] for par in mod_qm_mm.get_params() }
                d_par.update(d_par_bm_mm)
                d_par.update(d_par_qm_mm)

            if mod_ee is not None:
                d_par_ee = { par.name : [ par.value().numpy(), 0] for par in mod_ee.get_params() }
                d_par.update(d_par_ee)

        d_par['rk'] = [self._rk, 0]
        d_par       = self._add_ext_constraints(d_par, d_var)
        d_par       = self._add_ck(d_par)
        d_par       = self._add_ck_constraints(d_par, ck_var)

        return d_par
    #----------------------------------------------------
    def get_data(self, rseed=3):
        '''
        Returns toy data from model

        Parameters:
        -----------------
        rseed  (int):  Random seed

        Returns:
        -----------------
        d_data (dict): Dictionary with dataset and tuple of zfit data objects paired, i.e. {r1_TOS : (zdata_mm, zdata_ee) }

        For muon, TIS dataset is the TOS one.
        '''
        self._initialize()

        zfit.settings.set_seed(rseed)

        d_data     = {}
        zdat_tos_mm= None
        for ds, (pdf_mm, pdf_ee) in self._d_mod.items():
            zdat_mm = self._get_zdata(pdf_mm, ds, 'mm')
            zdat_ee = self._get_zdata(pdf_ee, ds, 'ee')

            if 'TIS' in ds:
                zdat_mm     = zdat_tos_mm 
            else:
                zdat_tos_mm = zdat_mm 

            d_data[ds] = zdat_mm, zdat_ee

        return d_data
    #----------------------------------------------------
    def _get_zdata(self, pdf, ds, chan):
        '''
        Will take either a PDF or a list of PDFs (muon channel), 
        Will return a zfit dataset or a list of zfit datasets.
        '''
        if pdf is None:
            return

        if isinstance(pdf, list):
            l_mod = pdf
        else:
            l_mod = [pdf]

        l_zdat = []
        for mod in l_mod:
            dst  = mod.create_sampler()
            arr  = dst.value().numpy().flatten()
            zdat = zfit.Data.from_numpy(mod.space, array=arr)
            l_zdat.append(zdat)

        if not isinstance(pdf, list):
            return l_zdat[0]

        return l_zdat
    #----------------------------------------------------
    @staticmethod
    def show(d_mod):
        s_dset = { key.split('_')[0] for key in d_mod }
        for dset in s_dset:
            pdf_mm_tos, pdf_ee_tos = d_mod[f'{dset}_TOS']
            pdf_mm_tis, pdf_ee_tis = d_mod[f'{dset}_TIS']

            l_par_name_mm_tos = ', '.join([ par.name for par in pdf_mm_tos.get_params() ])
            l_par_name_mm_tis = ', '.join([ par.name for par in pdf_mm_tis.get_params() ])
            l_par_name_ee_tos = ', '.join([ par.name for par in pdf_ee_tos.get_params() ])
            l_par_name_ee_tis = ', '.join([ par.name for par in pdf_ee_tis.get_params() ])

            log.info('')
            log.info(f'{dset}')
            log.info('-' * 20)
            log.info(f'{"mm TOS":<10}{l_par_name_mm_tos:<60}')
            log.info(f'{"mm TIS":<10}{l_par_name_mm_tis:<60}')
            log.info(f'{"ee TOS":<10}{l_par_name_ee_tos:<60}')
            log.info(f'{"ee TIS":<10}{l_par_name_ee_tis:<60}')
            log.info('-' * 20)
    #----------------------------------------------------
    @staticmethod
    def get_cov(kind='diag_eq', c = 0.01):
        if   kind == 'diag_eq':
            mat = numpy.diag([c] * 8)
        elif kind == 'random':
            mat = numpy.random.rand(8, 8)
            numpy.fill_diagonal(mat, 1)
            mat = mat * c
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return mat 
    #----------------------------------------------------
    @staticmethod
    def get_rjpsi(kind='one'):
        d_rjpsi = {}
    
        if   kind == 'one':
            d_rjpsi['d1'] = 1 
            d_rjpsi['d2'] = 1 
            d_rjpsi['d3'] = 1 
            d_rjpsi['d4'] = 1 
        elif kind == 'eff_bias':
            d_rjpsi['d1'] = 0.83333333 
            d_rjpsi['d2'] = 0.83333333 
            d_rjpsi['d3'] = 0.83333333 
            d_rjpsi['d4'] = 0.83333333 
        else:
            log.error(f'Wrong kind: {kind}')
            raise
    
        return d_rjpsi
    #----------------------------------------------------
    @staticmethod
    def get_eff(kind='equal'):
        d_eff = {}
        if   kind == 'diff':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.5, 0.2)
            d_eff['d3'] = (0.7, 0.3)
            d_eff['d4'] = (0.8, 0.4)
        elif kind == 'half':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.6, 0.3)
            d_eff['d3'] = (0.6, 0.3)
            d_eff['d4'] = (0.6, 0.3)
        elif kind == 'equal':
            d_eff['d1'] = (0.3, 0.3)
            d_eff['d2'] = (0.3, 0.3)
            d_eff['d3'] = (0.3, 0.3)
            d_eff['d4'] = (0.3, 0.3)
        elif kind == 'bias':
            d_eff['d1'] = (0.6, 0.25)
            d_eff['d2'] = (0.6, 0.25)
            d_eff['d3'] = (0.6, 0.25)
            d_eff['d4'] = (0.6, 0.25)
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return d_eff
#----------------------------------------------------




import os
import re
import glob
import ROOT
import math
import zfit
import numpy
import extset
import utils_noroot      as utnr
import zutils.utils      as zut
import matplotlib.pyplot as plt

from rdf_loader    import rdf_loader
from zutils.utils  import zfsp_1d_input
from misid_check   import misid_check
from zutils.utils  import split_fit     as zfsp
from zutils.plot   import plot          as zfp
from fitter        import zfitter
from log_store     import log_store

log = log_store.add_logger(name='rk_extractor:normalizer')
#--------------------------------------
class normalizer:
    '''
    Class used to find normalizations for combinatorial and PRec components
    in the high-q2 signal model
    '''
    #--------------------------------------
    def __init__(self, dset=None, trig=None, d_model=None, d_val=None, d_var=None, blind=False):
        self._d_model = d_model 
        self._dset    = dset
        self._trig    = trig
        self._d_val   = d_val
        self._d_var   = d_var
        self._blind   = blind

        self._l_flt_par= ['lm_cb', 'mu_cb', 'a', 'b', 'ncb', 'npr_ee','ncnd_', 'nsg_mm', 'cmb_sp_lam', 'ctrl_sp_sigma', 'nsg_ee', 'ctrl_sp_mu']
        self._l_scl_par= ['_dmu_', '_ssg_']

        self._d_const  = {}
        self._nbins    = 60 
        self._rng_mm   = 5180, 5600 
        self._rng_ee   = 4500, 6000 
        self._bln_ee   = 5000, 5450
        self._bln_mm   = 5200, 5360

        self._d_data      = None
        self._s_par       = None
        self._d_pre       = None
        self._out_dir     = None
        self._custom_data = False
        self._initialized = False 
    #--------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        os.makedirs(f'{self._out_dir}/pdf', exist_ok=True)

        self._check_none(self._d_model, 'self._d_model')
        self._check_none(self._d_val  , 'self._d_val'  )
        self._check_none(self._d_var  , 'self._d_var'  )

        self._s_par     = self._get_parameters() 
        self._d_pre     = {par.name : par.value().numpy() for par in self._s_par}
        self._d_const   = self._prepare_pars()
        self._d_data    = self._get_data()

        self._initialized = True 
    #--------------------------------------
    def _check_none(self, obj, name):
        if obj is None:
            log.error(f'Object {name} is None')
            raise
    #--------------------------------------
    def _get_parameters(self):
        s_par = set()
        for l_pdf_mm, pdf_ee in self._d_model.values():
            if l_pdf_mm is not None:
                [pdf_bm, pdf_qm] = l_pdf_mm
            else:
                [pdf_bm, pdf_qm] = [None, None] 

            s_mod_par_bm = set() if pdf_bm is None else pdf_bm.get_params()
            s_mod_par_qm = set() if pdf_qm is None else pdf_qm.get_params()
            s_mod_par_ee = set() if pdf_ee is None else pdf_ee.get_params()

            s_par        = s_par.union(s_mod_par_bm)
            s_par        = s_par.union(s_mod_par_qm)
            s_par        = s_par.union(s_mod_par_ee)

        return s_par
    #--------------------------------------
    @property
    def data(self):
        return self._d_data

    @data.setter
    def data(self, value):
        self._custom_data = True
        self._d_data      = value 
    #--------------------------------------
    def _container_tolist(self, arr_mass):
        if isinstance(arr_mass, zfit.Data):
            arr_mass = arr_mass.to_numpy()

        try:
            l_mass = arr_mass.tolist()
        except:
            log.error(f'Cannot convert array to list for:')
            log.info(arr_mass)
            raise

        return l_mass
    #--------------------------------------
    def _get_jpsi_mass(self, rdf):
        df  = misid_check.rdf_to_df(rdf, '(L1|L2|H)_(P\w|ID)$')
        obj = misid_check(df, d_lep={'L1' : 13, 'L2' : 13}, d_had={'H' : 13})
        df  = obj.get_df()

        arr_mass = df.H_swp.to_numpy()

        return arr_mass
    #-------------------------------------
    def _get_mass(self, rdf, kind):
        if   kind == 'bp':
            arr_mass = rdf.AsNumpy(['B_M'])['B_M']
        elif kind == 'mm':
            arr_mass = self._get_jpsi_mass(rdf)
        else:
            log.error(f'Invalid kind: {kind}')
            raise

        return arr_mass.tolist()
    #--------------------------------------
    @utnr.timeit
    def _get_data(self):
        dat_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.json'
        if   self._d_data is not None:
            log.warning(f'Using data passed by user')
            return self._d_data
        elif os.path.isfile(dat_path):
            log.info(f'Loading cached data from: {dat_path}')
            d_dat = utnr.load_json(dat_path)
            return d_dat 

        log.info(f'Caching data: {dat_path}')

        os.makedirs(f'{self._out_dir}/data', exist_ok=True)

        obj              = rdf_loader(sample='blind_fits', proc='data', asl_vers=None, ntp_vers='v10.21p2', year=self._dset, trig=self._trig)
        rdf, df_cf, d_md = obj.get_rdf()

        arr_bp = self._get_mass(rdf, 'bp')
        arr_mm = self._get_mass(rdf, 'mm') 
        d_dat  = {'bp_mass' : arr_bp, 'mm_mass' : arr_mm}

        log.info(f'Saving to: {dat_path}')
        utnr.dump_json(d_dat, dat_path) 

        cf_path = f'{self._out_dir}/data/{self._dset}_{self._trig}_cf.json'
        df_cf.to_json(cf_path, indent=4)

        md_path = f'{self._out_dir}/data/{self._dset}_{self._trig}_md.json'
        utnr.dump_json(d_md, md_path) 

        return d_dat 
    #--------------------------------------
    def _prepare_pars(self):
        for par in self._s_par:
            par.floating = False

        d_const = {}
        for par in self._s_par:
            for scl_par in self._l_scl_par:
                if scl_par in par.name:
                    par.floating = True
                    continue

            for flt_par in self._l_flt_par:
                if par.name.startswith(flt_par):
                    par.floating = True
                else:
                    continue

                if par.name in self._d_val:
                    var = self._d_var[par.name]
                    val = self._d_val[par.name]
                    d_const[par.name] = val, math.sqrt(var)

        return d_const
    #--------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create directory: {value}')
            raise

        self._out_dir = value
    #--------------------------------------
    def _get_stats(self, zdata, model):
        if self._trig != 'MTOS':
            return ''

        return 'no stats'

        if   isinstance(zdata, (zfit.core.data.SamplerData, zfit.core.data.Data)):
            arr_dat = zdata.numpy()
        elif isinstance(zdata, numpy.ndarray):
            arr_dat = zdata
        else:
            dat_typ = str(type(zdata))
            log.error(f'Cannot get a numpy array from: {dat_typ}')
            raise

        obs     = model.space
        [[minx]], [[maxx]] = obs.limits

        arr_flg = (arr_dat > minx) & (arr_dat < maxx)
        arr_dat = arr_dat[arr_flg]
        ndata   = float(arr_dat.size)

        s_par     = model.get_params(floating=False)
        [pst_val] = [ par.value().numpy() for par      in s_par               if par.name.startswith('nsg_mm_') ]
        [pre_val] = [ val                 for nam, val in self._d_pre.items() if nam.startswith('nsg_mm_')      ]

        v1 = f'Data: {ndata:.0f} $m\in [{minx:.0f}, {maxx:.0f}]$'
        v2 = f'Fitted: {pst_val:.0f}'
        v3 = f'Expected: {pre_val:.0f}'

        return f'{v1}\n{v2}\n{v3}'
    #--------------------------------------
    def _get_pdf_names(self):
        d_leg         = {}
        d_leg['prc']  = r'$c\bar{c}_{prc} + \psi(2S)K^+$'
        d_leg['bpks'] = r'$B^+\to K^{*+}(\to K^+\pi^0)e^+e^-$'
        d_leg['bdks'] = r'$B^0\to K^{*0}(\to K^+\pi^-)e^+e^-$'
        d_leg['bsph'] = r'$B_s\to \phi(\to K^+K^-) e^+e^-$'
        d_leg['bpk1'] = r'$B^+\to K_{1}e^+e^-$'
        d_leg['bpk2'] = r'$B^+\to K_{2}e^+e^-$'

        return d_leg
    #--------------------------------------
    def _get_blinding(self):
        if self._trig == 'MTOS':
            return

        low, hig= self._bln_ee
        l_blnd  = [extset.sig_name, low, hig]

        return l_blnd
    #--------------------------------------
    def _plot_fit(self, data=None, model=None, name=None, result=None, stacked=None):
        l_blnd  = self._get_blinding()
        stats   = self._get_stats(data, model)

        obj= zfp(data=data, model=model, result=None)
        obj.plot(skip_pulls=False, blind=l_blnd, nbins=self._nbins, d_leg=self._get_pdf_names(), stacked=stacked, ext_text=stats)
        obj.axs[0].grid()

        if not self._custom_data:
            if   '_all_' in name and 'MTOS' in name and '_bms_' in name:
                obj.axs[0].set_ylim(0,  250)
            elif '_all_' in name and 'MTOS' in name and '_bms_' in name:
                obj.axs[0].set_ylim(0,   50)
            elif '_all_' in name and 'ETOS' in name:
                obj.axs[0].set_ylim(0,  120)
            elif '_all_' in name and 'GTIS' in name:
                obj.axs[0].set_ylim(0,   50)

        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        vers  = os.path.basename(self._out_dir)
        title = f'Dataset: {self._dset}; Trigger: {self._trig}; Version: {vers}'
        obj.axs[0].set_title(title)

        os.makedirs(f'{self._out_dir}/fits', exist_ok=True)
        plot_path = f'{self._out_dir}/fits/{name}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def _reformat_mod(self, d_mod):
        l_mod_mm, mod_ee = d_mod['all']

        if   self._trig == 'MTOS':
            return l_mod_mm 
        elif self._trig in ['ETOS', 'GTIS']:
            return [mod_ee]
        else:
            log.error(f'Invalid trigger: {self._trig}')
            raise
    #--------------------------------------
    def _reformat_dat(self, d_dat):
        l_bm = d_dat['bp_mass']
        l_q2 = d_dat['mm_mass']

        arr_bm = numpy.array(l_bm)
        arr_q2 = numpy.array(l_q2)

        if   self._trig == 'MTOS':
            return [arr_bm, arr_q2]
        elif self._trig in ['ETOS', 'GTIS']:
            return [arr_bm]
        else:
            log.error(f'Invalid trigger: {self._trig}')
            raise
    #--------------------------------------
    @utnr.timeit
    def _fit(self, d_mod, d_dat):
        d_fdat  = {}
        d_fmod  = {}
        l_mod   = self._reformat_mod(d_mod)
        l_dat   = self._reformat_dat(d_dat)
        tot_nll = None
        for mod, dat in zip(l_mod, l_dat):
            zdat = zfit.Data.from_numpy(obs=mod.space, array=dat)
            name = mod.name

            d_fdat[name] = dat
            d_fmod[name] = mod

            if self._blind:
                log.info(f'Adding nll for model {name}, blinding')
                nll_1 = zfit.loss.ExtendedUnbinnedNLL(model=mod, data=zdat, fit_range=(4500, 5000))
                nll_2 = zfit.loss.ExtendedUnbinnedNLL(model=mod, data=zdat, fit_range=(5450, 6000))
                nll   = nll_1 + nll_2
            else:
                log.info(f'Adding nll for model {name}, unblinding')
                nll   = zfit.loss.ExtendedUnbinnedNLL(model=mod, data=zdat, fit_range=None)

            tot_nll = nll if tot_nll is None else nll + tot_nll

        minimizer = zfit.minimize.Minuit()
        log.debug('Minimizing')
        result    = minimizer.minimize(tot_nll)

        return result, d_fdat, d_fmod
    #--------------------------------------
    def _print_pdfs(self, prefix):
        for key, (l_mod_mm, mod_ee) in self._d_model.items():
            if l_mod_mm is not None:
                [mod_bm, mod_qm] = l_mod_mm
                zut.print_pdf(mod_bm, txt_path=f'{self._out_dir}/pdf/{prefix}_mm_{key}_{self._dset}_{self._trig}_bm.txt', d_const=self._d_const)
                zut.print_pdf(mod_qm, txt_path=f'{self._out_dir}/pdf/{prefix}_mm_{key}_{self._dset}_{self._trig}_qm.txt', d_const=self._d_const)

            if mod_ee is not None:
                zut.print_pdf(mod_ee, txt_path=f'{self._out_dir}/pdf/{prefix}_ee_{key}_{self._dset}_{self._trig}.txt', d_const=self._d_const)
    #--------------------------------------
    def _plot_fits(self, d_fdat, d_fmod):
        for key in d_fdat:
            arr_mass = d_fdat[key]
            model    = d_fmod[key]

            self._plot_fit(data=arr_mass, model=model, name=f'{self._trig}_{self._dset}_{key}_stk', stacked= True)
    #--------------------------------------
    @utnr.timeit
    def get_fit_result(self):
        self._initialize()

        self._print_pdfs('pre')
        res, d_fdat, d_fmod = self._fit(self._d_model, self._d_data)
        self._print_pdfs('pos')

        self._plot_fits(d_fdat, d_fmod)

        res.hesse()
        res.freeze()

        return res
#--------------------------------------




from model_analyzer import analyzer  as mana
from np_reader      import np_reader as np_rdr
from logzero        import logger    as log
from rk_model       import rk_model

import rk.utilities as rkut
import math
import zfit

#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#---------------------------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#---------------------------------------------
def get_model():
    obs          = zfit.Space('x', limits=(-10, 10))
    mu           = zfit.Parameter("mu", 2.4, -1, 5)
    sg           = zfit.Parameter("sg", 1.3,  0, 5)
    ne           = zfit.Parameter('ne', 100, 0, 1000)
    gauss        = zfit.pdf.Gauss(obs=obs, mu=mu, sigma=sg)

    return gauss.create_extended(ne) 
#---------------------------------------------
def test_fit():
    model = get_model()
    
    obj                    = mana(pdf=model, d_const={'mu' : [2.4, 0.5]})
    obj.out_dir            = 'tests/model_analyzer/pulls'
    df_ini, df_val, df_err = obj.fit(nfit=10)

    delete_all_pars()
#---------------------------------------------
def get_rk_model(kind):
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod   = rk_model(preffix='perf', d_eff=d_eff, d_nent=d_nent, l_dset=[kind])
    d_mod = mod.get_model()

    return d_mod[kind]
#---------------------------------------------
def get_cb_model(preffix, obs):
    mu = zfit.Parameter(f'{preffix}_mu', 5200, 5000, 5600)
    sg = zfit.Parameter(f'{preffix}_sg',   10,  0.1,  500)

    al = zfit.Parameter(f'{preffix}_al', 1, 0,  20)
    nl = zfit.Parameter(f'{preffix}_nl', 1, 0, 150)

    ar = zfit.Parameter(f'{preffix}_ar', 1, 0,  20)
    nr = zfit.Parameter(f'{preffix}_nr', 1, 0, 120)

    nev= zfit.Parameter(f'{preffix}_nev', 1000, 0, 2000)
    dscb = zfit.pdf.DoubleCB(
        mu    = mu,
        sigma = sg,
        alphal= al,
        nl    = nl,
        alphar= ar,
        nr    = nr,
        obs   = obs,
    )    

    pdf = dscb.create_extended(nev)

    return pdf 
#---------------------------------------------
def test_perf_sum():
    obs= zfit.Space('obs', limits=(5000, 5600))
    m1 = get_cb_model('m1', obs)
    m2 = get_cb_model('m2', obs)
    m3 = get_cb_model('m3', obs)
    sm = zfit.pdf.SumPDF([m1, m2, m3])

    obj         = mana(pdf=sm) 
    obj.out_dir = 'tests/model_analyzer/sum_perf'
    obj.sampling_speed(nsample=10)

    delete_all_pars()
#---------------------------------------------
def test_perf_rk():
    _, model = get_rk_model('all_TOS')

    obj         = mana(pdf=model) 
    obj.out_dir = 'tests/model_analyzer/rk_perf'
    obj.sampling_speed(nsample=10)

    delete_all_pars()
#---------------------------------------------
def test_rk():
    log.info('Getting nuisance parameters')
    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    rdr.cache_dir= './v65_v63_v24'
    d_eff        = rdr.get_eff()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    log.info('Building model')
    mod         = rk_model(preffix='all_tos', d_eff=d_eff, d_nent=d_rare_yld, l_dset=['all_TOS'])
    mod.bdt_bin = 2 
    d_mod       = mod.get_model()
    d_val, d_var= mod.get_cons()
    _, mod_ee   = d_mod['all_TOS']

    d_const = { key : [val, math.sqrt(var)] for key, val, var in zip(d_val, d_val.values(), d_var.values())}

    log.info('Analyzing model')
    obj            = mana(pdf=mod_ee, d_const = d_const, nev_fac=1)
    obj.out_dir    = 'tests/model_analyzer/rk'
    obj.sampling_speed(nsample=100)

    delete_all_pars()
#---------------------------------------------
def main():
    test_rk()
    return
    test_perf_sum()
    test_fit()
#---------------------------------------------
if __name__ == '__main__':
    main()




from importlib.resources import files
from datetime            import datetime
from log_store           import log_store
from version_management  import get_last_version 
from acceptance_reader   import reader            as acreader

import os
import re
import math
import glob
import ROOT
import numpy
import utils
import mplhep
import pprint
import pandas            as pnd
import jacobi            as jac
import utils_noroot      as utnr
import matplotlib.pyplot as plt

log=log_store.add_logger(name='rk_extractor:scales')
#------------------------------------------
def get_proc_labels():
    d_proc_lab         = dict() 
    d_proc_lab['bpkp'] = r'$B^+\to K^+ e^+e^-$'
    d_proc_lab['bpks'] = r'$B^+\to K^{*+}(\to K\pi^0) e^+e^-$'
    d_proc_lab['bdks'] = r'$B_d\to K^{*0}(\to K\pi) e^+e^-$'
    d_proc_lab['bsph'] = r'$B_s\to \phi(\to K K) e^+e^-$'
    d_proc_lab['bpk1'] = r'$B^+\to K_1(\to K \pi\pi) e^+e^-$'
    d_proc_lab['bpk2'] = r'$B^+\to K_2(\to K \pi\pi) e^+e^-$'

    return d_proc_lab
#------------------------------------------
class eff_calc:
    def __init__(self, proc=None, year=None, trig=None):
        self._l_proc = ['bpkp', 'bpks', 'bdks', 'bsph']                  if proc is None else proc
        self._l_trig = ['ETOS']                                          if trig is None else trig
        self._l_year = ['2011', '2012', '2015', '2016', '2017', '2018']  if year is None else year

        self._d_proc_lab = get_proc_labels()
        self._d_geo_eff  = {'Process' : [], 'Year' : [], 'Value' : [], 'Error' : []}
        self._d_sel_tis  = {'Process' : [], 'Year' : [], 'Value' : [], 'Error' : []}
        self._d_sel_tos  = {'Process' : [], 'Year' : [], 'Value' : [], 'Error' : []}
        self._dvers      = 'v10.21p2'
        self._min_date   = '01.09.23'
        self._out_dir    = None

        plt.style.use(mplhep.style.LHCb2)

        self._initialized=False
    #------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot make: {value}')
            raise

        self._out_dir = value
    #------------------------------------------
    def get_efficiencies(self):
        '''
        Parameters
        ---------------------
        '''
        d_eff = {}
        log.info('-' * 20)
        log.info(f'{"Process":<15}{"Trigger":<15}{"Year":<15}')
        log.info('-' * 20)
        for proc in self._l_proc: 
            d_eff[proc] = {}
            for trig in self._l_trig: 
                d_eff[proc][trig] = {}
                for year in self._l_year: 
                    log.info(f'{proc:<15}{trig:<15}{year:<15}')
                    d_eff[proc][trig][year] = self._get_efficiency(proc=proc, trig=trig, year=year)
        log.info('-' * 20)

        self._plot_geo_eff()
        self._plot_sel_eff()

        return d_eff
    #------------------------------------------
    def _plot_geo_eff(self):
        if self._out_dir is None:
            return

        df      = pnd.DataFrame(self._d_geo_eff)
        plt_dir = f'{self._out_dir}/plots'
        os.makedirs(plt_dir, exist_ok=True)

        ax=None
        for proc, df_p in df.groupby('Process'):
            ax=df_p.plot(x='Year', y='Value', yerr='Error', ax=ax, label=self._d_proc_lab[proc], marker='o')

        plt_path = f'{plt_dir}/geo_eff.png'
        log.info(f'Saving: {plt_path}')
        plt.grid()
        plt.ylim(0.00, 0.20)
        plt.savefig(plt_path)
        plt.close('all')
    #------------------------------------------
    def _plot_sel_eff(self):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots'
        os.makedirs(plt_dir, exist_ok=True)
        for trig, d_data in [('ETOS' , self._d_sel_tos)]:
            if trig not in self._l_trig:
                continue
            df = pnd.DataFrame(d_data) 

            ax=None
            for proc, df_p in df.groupby('Process'):
                ax=df_p.plot(x='Year', y='Value', yerr='Error', ax=ax, label=self._d_proc_lab[proc], marker='o')

            plt_path = f'{plt_dir}/sel_eff_{trig}.png'
            plt.grid()
            plt.savefig(plt_path)
            plt.close('all')
    #------------------------------------------
    def _switch_sample(self, proc, year):
        '''
        Switch samples for unavailable sim08 generator efficiencies
        '''
        if  proc in ['bpk1_ee', 'bpk2_ee'] and year in ['2011', '2012', '2015', '2016']:
            new_year = '2017'
            log.warning(f'Gen efficiency using {year} -> {new_year} for {proc}')
            year = new_year 

        return proc, year
    #------------------------------------------
    def _get_geo_eff(self, proc, year):
        obj = acreader(year=year, proc=proc)
        acc = obj.read()

        self._d_geo_eff['Process'].append(proc) 
        self._d_geo_eff['Year'   ].append(year) 
        self._d_geo_eff['Value'  ].append( acc) 
        self._d_geo_eff['Error'  ].append(   0) 

        return acc
    #------------------------------------------
    def _get_efficiency(self, proc=None, trig=None, year=None):
        sel_eff  = self._get_sel_eff(proc, trig, year)
        geo_acc  = self._get_geo_eff(proc,       year)
        geo_acc  = [geo_acc, 0]
        log.warning(f'Taking geometric acceptance error as zero')

        val, var = jac.propagate(lambda x : x[0] * x[1], [geo_acc[0], sel_eff[0]], [[geo_acc[1] ** 2, 0], [0, sel_eff[1] ** 2]]) 
        val      = float(val)
        err      = math.sqrt(var)
        d_data   = self._d_sel_tos if trig == 'ETOS' else self._d_sel_tis

        d_data['Process'].append(proc)
        d_data['Year'   ].append(year)
        d_data['Value'  ].append( val)
        d_data['Error'  ].append( err)

        return val, err
    #------------------------------------------
    def _get_sel_yield(self, proc, trig, year):
        proc     = 'sign' if proc == 'bpkp' else proc
        cas_dir  = os.environ['CASDIR']
        root_wc  = f'{cas_dir}/tools/apply_selection/rare_backgrounds/{proc}/{self._dvers}/{year}_{trig}/*.root'

        l_root   = glob.glob(root_wc)
        if len(l_root) == 0:
            log.error(f'No file found in: {root_wc}')
            raise

        rdf = ROOT.RDataFrame(trig, l_root)

        return rdf.Count().GetValue()
    #------------------------------------------
    def _get_date_flag(self, date):
        l_date = re.findall('\d{2}\.\d{2}\.\d{2}', date)
        if len(l_date) == 0:
            return False

        dstr = datetime.strptime

        return any( dstr(self._min_date, '%d.%m.%y') < dstr(date, '%d.%m.%y') for date in l_date)
    #------------------------------------------
    def _switch_yields(self, proc, year):
        if   proc in ['bpk1', 'bpk2'] and year in ['2015', '2016']:
            new_year = '2017'
            log.warning(f'Reco yields using {year} -> {new_year} for {proc}')
            year = new_year
        elif proc == 'bpk2' and year == '2011':
            new_year = '2012'
            log.warning(f'Reco yields using {year} -> {new_year} for {proc}')
            year = new_year

        return proc, year
    #------------------------------------------
    def _get_year_entries(self, df, proc, year):
        proc, year = self._switch_yields(proc, year)

        df_f = df
        df_f = df_f[(df_f.Year     == str(year)) | (df_f.Year     == int(year))]
        df_f = df_f[(df_f.Polarity == 'MagUp'  ) | (df_f.Polarity == 'MagDown')]

        if len(df_f) not in [1, 2]:
            log.error(f'Found more than two or fewer than one entries (polarities) after fully filtering')
            print(df)
            print(df.dtypes)
            log.info('--->')
            print(df_f)
            raise

        l_pol=df_f.Polarity.tolist()
        s_pol=set(l_pol)
        if   len(df_f) == 2 and s_pol != {'MagUp', 'MagDown'}:
            log.error(f'Wrong polarities: {s_pol}')
            raise ValueError
        elif len(df_f) == 1:
            log.warning(f'Found only polarity: {s_pol}')

        return df_f.Events.sum()
    #------------------------------------------
    def _get_gen_yield(self, proc, year):
        ganga_dir = os.environ['GANDIR']
        json_path = f'{ganga_dir}/job_stats/{proc}.json'

        df = pnd.read_json(json_path)
        df = df.loc[[ self._get_date_flag(date) for date in df.Dates ]]

        yld= self._get_year_entries(df, proc, year)

        return yld 
    #------------------------------------------
    def _get_sel_eff(self, proc, trig, year):
        sel = self._get_sel_yield(proc, trig, year)
        gen = self._get_gen_yield(proc,       year)

        eff, eup, edn = utils.get_eff_err(sel, gen)

        err = (eup + edn) / 2.

        return eff, err
#------------------------------------------
class scales:
    def __init__(self, dset=None, trig=None, kind=None):
        self._dset = dset
        self._trig = trig 
        self._kind = kind 

        self._l_kind = ['bpks', 'bdks', 'bsph', 'bpk1', 'bpk2']
        self._l_trig = ['ETOS', 'GTIS']
        self._l_dset = ['2011', '2012', '2015', '2016', '2017', '2018']

        self._d_frbf = None
        self._d_eff  = None

        self._initialized = False
    #------------------------------------------
    def _check_arg(self, l_val, val, name):
        if val not in l_val:
            log.error(f'{name} {val} not allowed')
            raise ValueError
        else:
            log.debug(f'{name:<20}{"->":20}{val:<20}')
    #------------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        log.debug('Initializing')
        self._vers = self._get_fractions_version()

        self._load_fractions()
        self._load_efficiencies() 

        self._check_arg(self._l_kind, self._kind, 'Background')
        self._check_arg(self._l_dset, self._dset, 'Dataset')
        self._check_arg(self._l_trig, self._trig, 'Trigger')

        self._initialized = True
    #------------------------------------------
    def _get_fractions_version(self):
        frc_dir = files('extractor_data').joinpath(f'rare_sf')
        vers    = get_last_version(dir_path=frc_dir, version_only=True)

        log.debug(f'Using fractions version: {vers}')

        return vers
    #------------------------------------------
    def _load_fractions(self):
        log.debug(f'Using rare-signal scales version: {self._vers}')
        frbf_path    = files('extractor_data').joinpath(f'rare_sf/{self._vers}/fr_bf.json')
        self._d_frbf = utnr.load_json(frbf_path)
    #------------------------------------------
    def _load_efficiencies(self):
        log.debug('Getting efficiencies')

        eff_path     = files('extractor_data').joinpath(f'rare_sf/{self._vers}/eff_real.json'  )
        if not os.path.isfile(eff_path):
            log.debug('Not found efficiencies file, remaking it')
            calc   = eff_calc()
            d_eff  = calc.get_efficiencies()
            utnr.dump_json(d_eff, eff_path)
            log.debug(f'Efficiencies file saved to: {eff_path}')
        else:
            d_eff  = utnr.load_json(eff_path)
            log.debug('Found efficiencies file, not remaking it')

        self._d_eff = d_eff
    #------------------------------------------
    def _get_fr(self, kind):
        key= 'bpkp' if kind == 'sig' else self._kind
        fx = {'bpkp' : 'fu', 'bpks' : 'fu', 'bdks' : 'fd', 'bsph' : 'fs', 'bpk1' : 'fu', 'bpk2' : 'fu'}[key]
        fx = self._d_frbf['fr'][fx]

        return fx
    #------------------------------------------
    def _mult_brs(self, br_1, br_2):
        l_br_val = [br_1[0], br_2[0]]
        br_cov   = [[br_1[1] ** 2, 0], [0, br_2[1] ** 2]]
        val, var = jac.propagate(lambda x : x[0] * x[1], l_br_val, br_cov)

        return val, math.sqrt(var)
    #------------------------------------------
    def _get_br(self, kind):
        if   kind == 'sig':
            br = self._d_frbf['bf']['bpkp']
            return br

        key = self._kind

        if   key in 'bpk1':
            b1 = self._d_frbf['bf'][   key]
            b2 = self._d_frbf['bf']['k13h']
        elif key in 'bpk2':
            b1 = self._d_frbf['bf'][   key]
            b2 = self._d_frbf['bf']['k23h']
        elif key == 'bpks':
            b1 = self._d_frbf['bf'][   key]
            b2 = self._d_frbf['bf']['k+kp']
        elif key == 'bdks':
            b1 = self._d_frbf['bf'][   key]
            b2 = self._d_frbf['bf']['kokp']
        elif key == 'bsph':
            b1 = self._d_frbf['bf'][   key]
            b2 = self._d_frbf['bf']['phkk']
        else:
            log.error(f'Invalid key: {key}')
            raise

        br = self._mult_brs(b1, b2)

        return br
    #------------------------------------------
    def _get_ef(self, kind):
        key = 'bpkp' if kind == 'sig' else self._kind

        return self._d_eff[key][self._trig][self._dset]
    #------------------------------------------
    def _print_vars(self, l_tup):
        log.debug('-' * 20)
        log.debug(f'{"Var":<20}{"Value":<20}{"Error":<20}')
        log.debug('-' * 20)
        for (val, err), name in zip(l_tup, ['fr sig', 'br sig', 'eff sig', 'fr bkg', 'br bkg', 'eff bkg']):
            log.debug(f'{name:<20}{val:<20.3e}{err:<20.3e}')
        log.debug('-' * 20)
    #------------------------------------------
    def get_scale(self):
        self._initialize()

        fr_sig = self._get_fr('sig') 
        br_sig = self._get_br('sig') 
        ef_sig = self._get_ef('sig') 

        fr_bkg = self._get_fr('bkg') 
        br_bkg = self._get_br('bkg') 
        ef_bkg = self._get_ef('bkg') 

        l_tup = [fr_sig, br_sig, ef_sig, fr_bkg, br_bkg, ef_bkg]
        l_val = [ tup[0] for tup in l_tup]
        l_err = [ tup[1] for tup in l_tup]
        cov   = numpy.diag(l_err) ** 2

        self._print_vars(l_tup)

        val, var = jac.propagate(lambda x : (x[3] * x[4] * x[5]) / (x[0] * x[1] * x[2]), l_val, cov) 
        val = float(val)
        err = math.sqrt(var)

        return val, err 
#------------------------------------------




from mc_reader import mc_reader as mc_rdr
import pprint
import pytest
import os

#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#-----------------------------
def test_simple():
    skip_test()
    rdr        = mc_rdr(version='v4')
    rdr.cache  = False 
    d_mcsg     = rdr.get_parameter(name='sg')
    d_mcmu     = rdr.get_parameter(name='mu')
#-----------------------------
def test_data():
    skip_test()
    rdr        = mc_rdr(version='v4', real_data=True)
    rdr.cache  = False 
    d_mcsg     = rdr.get_parameter(name='sg')
    d_mcmu     = rdr.get_parameter(name='mu')
#-----------------------------
def test_cache():
    skip_test()
    rdr        = mc_rdr(version='v4')
    rdr.cache  = True 
    rdr.cache_dir = 'tests/mc_reader/v4_mcrdr'
    d_mcsg     = rdr.get_parameter(name='sg')
    d_mcmu     = rdr.get_parameter(name='mu')
#-----------------------------
def main():
    test_data()
    test_simple()
    test_cache()
#-----------------------------
if __name__ == '__main__':
    main()




from log_store import log_store
import extset
import logzero
#log_store.set_level('rk_extractor:scale_reader', logzero.DEBUG)
log_store.set_level('rk_extractor:scale_reader', logzero.INFO)

from bdt_scale import scale_reader as scl_rdr
#------------------------------------------------
def test_zero():
    prcno            = 0.480751
    cmbno            = 0.831497

    d_wp             = {}
    d_wp['BDT_prc']  = prcno, 100 
    d_wp['BDT_cmb']  = 0.730, cmbno 

    obj = scl_rdr(wp=d_wp, version='v1', dset='r1', trig='ETOS')
    scl = obj.get_scale()

    print(scl)
#------------------------------------------------
def test_simple():
    prcno            = 0.480751
    cmbno            = 0.831497

    d_wp             = {}
    d_wp['BDT_prc']  = prcno, 100 
    d_wp['BDT_cmb']  = cmbno, 0.900 
    d_wp['BDT_cmb']  = 0.900, 0.977 
    d_wp['BDT_cmb']  = 0.977, 100 
    #d_wp['BDT_cmb']  = 0.800, None 

    obj = scl_rdr(wp=d_wp, version='v1', dset='r1', trig='ETOS')
    scl = obj.get_scale()

    print(scl)
#------------------------------------------------
def test_real():
    for bdt_bin in [1, 2, 3, 4, 5]:
        for dset in ['r1', 'r2p1', '2017', '2018', 'all']:
            for trig in ['MTOS', 'ETOS']:
                d_bdt_wp, _, _ = extset.get_bdt_bin_settings(bdt_bin = bdt_bin)
                obj = scl_rdr(wp=d_bdt_wp, version='v1', dset=dset, trig=trig)
                scl = obj.get_scale()
                print(scl)
#------------------------------------------------
def main():
    test_real()
    test_simple()
    test_zero()
#------------------------------------------------
if __name__ == '__main__':
    main()




import os
os.environ['ZFIT_DISABLE_TF_WARNINGS']='1'

import zfit
zfit.settings.changed_warnings.all = False

import logzero
from log_store import log_store

import zutils.utils as zut

log_store.set_level('rk_extractor:rk_model'      , logzero.INFO)
log_store.set_level('rk_extractor:rkex_model'    , logzero.INFO)
log_store.set_level('rk_extractor:scales'        , logzero.WARNING)
log_store.set_level('rk_extractor:scale_reader'  , logzero.WARNING)
log_store.set_level('rk_selection:read_selection', logzero.WARNING)
log_store.set_level('cb_calculator:builder'      , logzero.WARNING)
log_store.set_level('scripts:zutils/utils'       , logzero.INFO)

from model_manager import manager
from rk_model      import rk_model  as model

#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#----------------------------------------------------
def test_simple():
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e2, 'd2' :            2e2, 'd3' :            2e2, 'd4' :            2e2} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = manager(preffix='simple', d_eff=d_eff, d_nent=d_yld, dset='all_TOS')
    mod.fake    = True
    mod.bdt_bin = [1, 2]
    mod.com_par = ['mu_ee']
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    for _, (_, pdf) in d_mod.items():
        zut.print_pdf(pdf)
#----------------------------------------------------
def test_channel():
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e2, 'd2' :            2e2, 'd3' :            2e2, 'd4' :            2e2} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = manager(preffix='simple', d_eff=d_eff, d_nent=d_yld, dset='all_TOS', chan='mm')
    mod.fake    = False 
    mod.bdt_bin = [5]
    mod.com_par = ['mu_ee']
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    for _, (pdf, _) in d_mod.items():
        if pdf is None:
            continue 
        zut.print_pdf(pdf)
#----------------------------------------------------
def test_real():
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e2, 'd2' :            2e2, 'd3' :            2e2, 'd4' :            2e2} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = manager(preffix='simple', d_eff=d_eff, d_nent=d_yld, dset='all_TOS')
    mod.fake    = False 
    mod.bdt_bin = [4, 5]
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()
#----------------------------------------------------
def test_fit():
    d_eff       = model.get_eff()
    d_yld       = {'d1' :          2e2, 'd2' :            2e2, 'd3' :            2e2, 'd4' :            2e2} 
    d_mcmu      = {'d1' : (5000, 4900), 'd2' :   (5100, 4900), 'd3' :   (5100, 4800), 'd4' :   (5200, 5100)}
    d_mcsg      = {'d1' :    (10,  40), 'd2' :       (10, 18), 'd3' :       (20, 30), 'd4' :       (30, 40)} 

    d_eff       = rename_keys(d_eff)
    d_yld       = rename_keys(d_yld, use_txs=False)
    d_mcmu      = rename_keys(d_mcmu)
    d_mcsg      = rename_keys(d_mcsg)

    mod         = manager(preffix='fit', d_eff=d_eff, d_nent=d_yld, dset='all_TOS')
    mod.fake    = True
    mod.bdt_bin = [1, 2, 3, 4, 5]
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    sim_fit(d_dat, d_mod)
#----------------------------------------------------
def sim_fit(d_dat, d_mod):
    nll = None
    for key in d_dat:
        _, ee_dat = d_dat[key]
        _, ee_mod = d_mod[key]

        nll_ee = zfit.loss.ExtendedUnbinnedNLL(model=ee_mod, data=ee_dat)

        nll = nll_ee if nll is None else nll + nll_ee

    minimizer    = zfit.minimize.Minuit()
    result       = minimizer.minimize(nll)
    result.hesse()

    print(result)
#----------------------------------------------------
def main():
    test_channel()
    return
    test_simple()
    test_real()
    test_fit()
#----------------------------------------------------
if __name__ == '__main__':
    main()




from np_reader import np_reader as np_rdr
from log_store import log_store
import pprint
import os

log=log_store.add_logger('rk_extractor:test_npreader')
#-----------------------------
def test_simple():
    log.info('simple')
    rdr       = np_rdr(sys='v65', sta='v81', yld='v24')
    rdr.cache = True
    d_eff     = rdr.get_eff()
    cov_sys   = rdr.get_cov(kind='sys')
    cov_sta   = rdr.get_cov(kind='sta')
    df_rjpsi  = rdr.get_rjpsi()
    df_byld   = rdr.get_byields()
    df_ryld   = rdr.get_ryields()

    pprint.pprint(df_ryld , indent=4)
    pprint.pprint(df_rjpsi, indent=4)

    mtos = df_ryld.sign_MTOS.sum()
    etos = df_ryld.sign_ETOS.sum()

    print('MTOS', mtos)
    print('ETOS', etos)
#-----------------------------
def main():
    test_simple()
#-----------------------------
if __name__ == '__main__':
    main()




import ROOT
import zfit
import numpy
import math

from rk_model    import rk_model
from logzero     import logger    as log
from mc_reader   import mc_reader as mc_rdr
from np_reader   import np_reader as np_rdr
from cs_reader   import cs_reader as cs_rdr
from zutils.plot import plot      as zfp

import utils_noroot      as utnr
import matplotlib.pyplot as plt
import zutils.utils      as zut
import rk.utilities      as rkut
import pytest
import pprint
import os

#------------------------------------
def plot(shape, label, mass_window=(4500, 6000), d_const=None):
    plot_dir = f'tests/rk_model/{label}'
    os.makedirs(plot_dir, exist_ok=True)

    obj   = zfp(data=shape.arr_mass, model=shape)
    obj.plot(nbins=50, stacked=True)

    log.info(f"saving plot to {plot_dir}/pdf.png")
    plt.savefig(f'{plot_dir}/pdf.png')
    plt.close()

    zut.print_pdf(shape, d_const=d_const, txt_path=f'{plot_dir}/pdf.txt')
#-----------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename         = {}
    d_rename[  'r1'] = d_data['d1'], d_data['d1']
    d_rename['r2p1'] = d_data['d2'], d_data['d2']
    d_rename['2017'] = d_data['d3'], d_data['d3']
    d_rename['2018'] = d_data['d4'], d_data['d4']

    return d_rename
#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#----------------------
def get_syst():
    l_syst = []
    l_syst.append(('cpr_ETOS:bts1', 'cpr_etos_b1'))
    l_syst.append(('rpr_ETOS:bts1', 'rpr_etos_b1'))

    l_syst.append((          'nom',         'nom'))
    l_syst.append(('sig_ETOS:sys1', 'sig_etos_s1'))
    l_syst.append(('sig_MTOS:sys1', 'sig_mtos_s1'))
    l_syst.append(('cpr_ETOS:sys1', 'cpr_etos_s1'))

    return l_syst
#----------------------
def test_syst():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    l_syst = get_syst()
    for syst, lab in l_syst: 
        mod         = rk_model(preffix='syst', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
        mod.bdt_bin = 5
        mod.kind    = syst 
        mod.plt_set = {'stacked' : True, 'no_data' : True, 'ymax' : [170, 170] }
        mod.out_dir = f'tests/rk_model/syst/{lab}'
        d_mod       = mod.get_model()

        delete_all_pars()
#----------------------
def test_bootstrap():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='bootstrap', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
    mod.bdt_bin = 5
    mod.kind    = 'prc_rare:bts_001'
    mod.out_dir = 'tests/rk_model/bootstrap'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
@utnr.timeit
def test_bdt_change():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    for bdt_bin in [1, 2, 3, 4, 5]:
        mod         = rk_model(preffix=f'bdt_change_{bdt_bin}', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
        mod.bdt_bin = bdt_bin
        mod.out_dir = f'tests/rk_model/bdt_change/bdt_{bdt_bin}'
        d_mod       = mod.get_model()

        delete_all_pars()
#----------------------
def test_all_tos():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='all_tos', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
    d_mod       = mod.get_model()
    d_val, d_var= mod.get_cons()
    _, mod_ee   = d_mod['all_TOS']

    mod_ee.arr_mass = mod_ee.create_sampler(fixed_params=False)

    d_const = { key : [val, math.sqrt(var)] for key, val, var in zip(d_val, d_val.values(), d_var.values())}

    plot(mod_ee, 'all_tos', d_const=d_const)

    delete_all_pars()
#----------------------
def test_wp():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='wp_ap', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    mod.bdt_wp  = {'BDT_cmb' : 0.9, 'BDT_prc' : 0.7}
    mod.out_dir = 'tests/rk_model/wp/ap'
    d_mod       = mod.get_model()

    delete_all_pars()

    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='wp_no', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    mod.out_dir = 'tests/rk_model/wp/no'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_kind():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='kind', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    mod.kind    = 'cmb_ee:use_etos'
    mod.out_dir = 'tests/rk_model/kind'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_all_years():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='allyears', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
    mod.out_dir = 'tests/rk_model/all_years'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_dataset():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='data', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    mod.bdt_bin = 5
    d_zdat      = mod.get_data()

    delete_all_pars()
#----------------------
def test_rseed():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod_1   = rk_model(preffix='data1', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_dat_1 = mod_1.get_data(rseed=0)

    mod_2   = rk_model(preffix='data2', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_dat_2 = mod_2.get_data(rseed=0)

    plt_dir = f'tests/rk_model/rseed'
    os.makedirs(plt_dir, exist_ok=True)
    for index in [0, 1]:
        kind = 'mm' if index == 0 else 'ee'
        for key in d_dat_1:
            arr_dat_1 = d_dat_1[key][index].numpy().flatten()
            arr_dat_2 = d_dat_2[key][index].numpy().flatten()

            close = numpy.allclose(arr_dat_1, arr_dat_2, atol=1e-5)
            assert(close)

            min_x, max_x = min(arr_dat_1), max(arr_dat_1)

            plt.hist(arr_dat_1, bins=30, range=(min_x, max_x), histtype='step', linestyle='-', label='First')
            plt.hist(arr_dat_2, bins=30, range=(min_x, max_x), histtype='step', linestyle=':', label='Second')
            plt.legend()
            plt.savefig(f'{plt_dir}/{key}_{kind}.png')
            plt.close('all')

    delete_all_pars()
#----------------------
def test_cons():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)

    mod         = rk_model(preffix='cons', d_eff=d_eff, d_nent=d_nent, l_dset=['2018_TOS'])
    d_mod       = mod.get_cons()

    delete_all_pars()
#----------------------
@utnr.timeit
def test_channel():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e3, 'd2' :          1e3, 'd3' :          1e3, 'd4' :          1e3}
    d_eff = rename_keys(d_eff)
    d_nent= rename_keys(d_nent, use_txs=False)
    obs_mm= zfit.Space('mass mm', limits=(2600, 3300))

    mod = rk_model(
            preffix = 'channel', 
            d_eff   = d_eff, 
            d_nent  = d_nent, 
            l_dset  = ['all_TOS'], 
            channel = 'mm',
            obs_mm_sp = obs_mm,
            )

    mod.bdt_bin = 5
    mod.out_dir = 'tests/rk_model/channel'
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
@utnr.timeit
def test_channel_mm():
    d_nent= {'all' : (1.5e3, 300)}
    obs_mm= zfit.Space('mass mm', limits=(2600, 3300))

    mod = rk_model(
            preffix = 'channel', 
            d_nent  = d_nent, 
            channel = 'mm',
            obs_mm_sp = obs_mm,
            )

    mod.bdt_bin = 5
    mod.out_dir = 'tests/rk_model/channel_mm'
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
@utnr.timeit
def test_channel_ee():
    d_nent= {'all' : (1.5e3, 300)}
    obs_mm= zfit.Space('mass mm', limits=(2600, 3300))

    mod = rk_model(
            preffix = 'channel', 
            d_nent  = d_nent, 
            channel = 'ee',
            obs_mm_sp = obs_mm,
            )

    mod.bdt_bin = 5
    mod.out_dir = 'tests/rk_model/channel_ee'
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
@utnr.timeit
def test_simple():
    d_nent= {'all' : (1.5e3, 300)}
    obs_mm= zfit.Space('mass mm', limits=(2600, 3300))

    mod   = rk_model(
            preffix   = 'simple', 
            d_nent    = d_nent,
            obs_mm_sp = obs_mm,
            ) 

    mod.bdt_bin = 5
    mod.out_dir = 'tests/rk_model/simple'
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def main():
    utnr.timer_on = True
    test_simple()
    test_channel_ee()
    test_channel_mm()
    test_dataset()
    test_syst()
    test_bootstrap()
    test_bdt_change()
    test_all_tos()
    test_kind()
    test_wp()
    test_cons()
    test_rseed()
    test_all_years()
#----------------------
if __name__ == '__main__':
    main()




import re
import os
import zfit
import numpy
import pprint
import extset
import zutils.utils      as zut
import utils_noroot      as utnr
import hqm.model         as hqm_model
import matplotlib.pyplot as plt
import read_selection    as rs

from log_store             import log_store
from importlib.resources   import files
from rkex_model            import model
from version_management    import get_last_version
from zutils.utils          import zfsp_1d_input
from phsp_cmb.phsp_cmb     import cmb_pdf                  as phsp_cmb_pdf
from zutils.plot           import plot                     as zfp
from zutils.utils          import split_fit                as zfsp
from scales                import scales                   as scl
from stats.average         import average                  as stav
from misID_tools.zmodel    import misID_real_model_builder as msid
from hqm.model.double_swap import dswp_bmass               as dswpb
from builder               import builder                  as cb_builder

log = log_store.add_logger(name='rk_extractor:rk_model')
#---------------------------------------------------------------
class rk_model(model):
    def __init__(self, **kwargs):
        self._d_val      = {} 
        self._d_var      = {} 
        self._d_plt_set  = {} 

        self._obs        = None
        self._nsig_ee    = None
        self._nsig_mm    = None
        self._read_yields= True

        super().__init__(**kwargs)
    #---------------------------------------------------------------
    @property
    def plt_set(self):
        return self._d_plt_set

    @plt_set.setter
    def plt_set(self, value):
        self._d_plt_set = value
    #---------------------------------------------------------------
    @property
    def read_yields(self):
        return self._read_yields

    @read_yields.setter
    def read_yields(self, value):
        '''
        If true, will try to read yields from JSON files. This is needed to make toys
        not for actual fit to data
        '''
        self._read_yields= value
    #---------------------------------------------------------------
    def _preffix_to_info(self, preffix):
        regex = '(ee|mm)_(r1|r2p1|2017|2018|all)'
        mtch  = re.match(regex, preffix)
        try:
            [chan, dset] = mtch.groups()
        except:
            log.error(f'Cannot extract dataset and trigger from: {preffix}')
            raise

        trig = {'ee' : 'ETOS', 'mm' : 'MTOS'}[chan]

        return dset, trig
    #---------------------------------------------------------------
    def _add_constraints(self, d_cns):
        d_val = { name : cns[0]    for name, cns in d_cns.items() }
        d_var = { name : cns[1]**2 for name, cns in d_cns.items() }

        self._d_val.update(d_val)
        self._d_var.update(d_var)
    #---------------------------------------------------------------
    def _get_ph_combinatorial(self, preffix):
        _, trig = self._preffix_to_info(preffix) 
        qsq_cut = rs.get('q2', trig, q2bin='high', year='2018')
        regex   = '\(Jpsi_M \* Jpsi_M > (15500000.0)\) && \(Jpsi_M \* Jpsi_M < 22000000\.0\)'
        mtch    = re.match(regex, qsq_cut)
        if not mtch:
            log.error(f'Cannot extract minimum q2 from {qsq_cut} with {regex}')
            raise

        qmin = mtch.group(1)
        qmin = float(qmin)
        qmin = qmin / 1e6

        lam=zfit.param.Parameter(f'lam_{preffix}', 2, 1, 3)
        qmi=zfit.param.ConstantParameter(f'qmi_{preffix}', qmin)
        pdf=phsp_cmb_pdf(lam, qmi, self._obs)

        return pdf
    #---------------------------------------------------------------
    def _get_su_combinatorial(self, preffix):
        log.info(f'Getting combinatorial PDF for {preffix}')
        dset, trig    = self._preffix_to_info(preffix) 
        obj           = cb_builder(dset=dset, trigger=trig, version=self._comb_vers) 
        obj.kind      = self._kind 
        log.warning(f'Not using OS-SS correction, just SS to get combinatorial')
        cmb           = obj.get_pdf(obs=self._obs, preffix=f'cb_{preffix}', name='Combinatorial') 

        if 'sb_fits' not in preffix:
            log.warning('Fixing combinatorial shape, not using constraints')
            zut.fix_shape(cmb)
        else:
            log.debug('Not fixing shape of combinatorial')

        return cmb 
    #---------------------------------------------------------------
    def _get_combinatorial(self, preffix, fix_norm=False):
        dset, trig = self._preffix_to_info(preffix) 
        #pdf        = self._get_ph_combinatorial(preffix)
        pdf        = self._get_su_combinatorial(preffix)
        nent       = self._get_entries(nsig=None, kind='cmb', trigger=trig, year=dset)
        ncb        = zfit.Parameter(f'ncb_{preffix}', nent, 0, 100000)
        ncb.floating = not fix_norm 
        pdf.set_yield(ncb)

        return pdf
    #---------------------------------------------------------------
    def _get_rare_scale(self, year, trigger, kind):
        l_all_year = ['2011', '2012', '2015', '2016', '2017', '2018']

        if   year in l_all_year:
            obj      = scl(dset=year, trig=trigger, kind=kind)
            val, err = obj_1.get_scale()

            return val, err
        elif year == 'r1':
            l_year = ['2011', '2012']
        elif year == 'r2p1':
            l_year = ['2015', '2016']
        elif year == 'all':
            l_year = l_all_year 
        else:
            log.error(f'Invalid year: {year}')
            raise

        l_val = []
        l_err = []
        for dset in l_year: 
            obj      = scl(dset=dset, trig=trigger, kind=kind)
            val, err = obj.get_scale()
            l_val.append(val)
            l_err.append(err)

        arr_val     = numpy.array(l_val)
        arr_err     = numpy.array(l_err)
        val, err, _ = stav(arr_val, arr_err)

        return val, err
    #---------------------------------------------------------------
    def _get_entries_from_json(self, kind=None, trig=None, year=None):
        if not self._read_yields:
            log.warning(f'Not reading yields from JSON for {kind}/{trig}/{year}, using zero events')
            return 0

        json_dir  = files('extractor_data').joinpath(f'sb_fits')
        vers      = get_last_version(dir_path=json_dir, version_only=True)
        json_path = f'{json_dir}/{vers}/{year}_{trig}.json'
        if not os.path.isfile(json_path):
            log.error(f'Cannot find: {json_path}')
            raise FileNotFoundError

        d_par = utnr.load_json(json_path)
        for key, [val, _] in d_par.items():
            flg_1 = key.startswith('ncb_') and kind == 'cmb'
            flg_2 = key.startswith('npr_') and kind == 'prc'
            flg_3 = key == 'ncnd_ctrl_sp'  and kind == 'dswp'
            if flg_1 or flg_2 or flg_3:
                return val

        log.error(f'Cannot find yield for {kind} among:')
        pprint.pprint(d_par)
        raise
    #---------------------------------------------------------------
    def _get_entries(self, nsig=None, kind=None, trigger=None, year=None):
        if   kind in ['prc', 'cmb', 'dswp']:
            nent  = self._get_entries_from_json(kind=kind, trig=trigger, year=year)

            log.debug(f'For {trigger}/{year}, taking {kind} yield from JSON as: {nent:.0f}')
        elif kind in ['bpks', 'bdks', 'bsph', 'bpk1', 'bpk2']:
            scale = self._get_rare_scale(year, trigger, kind)
            scl   = scale[0]
            nent  = nsig * scl

            log.debug(f'For {trigger}/{year}, taking {kind} yield as {nent:.0f}={scl:.3f} * {nsig:.0f}')
        else:
            log.error(f'Invalid kind: {kind}')
            raise

        return nent
    #---------------------------------------------------------------
    def _get_systematic(self, trig, kind):
        if trig not in ['ETOS', 'MTOS']:
            log.error(f'Cannot pick up systematic for trigger {trig}')
            raise

        prefix = f'{kind}_{trig}:'
        if self._kind.startswith(prefix):
            syst = self._kind.replace(prefix, '')
            log.info(f'Using {syst} systematic for {kind}/{trig}')
        else:
            syst = 'nom'

        if   syst.startswith('sys') or syst == 'nom':
            return syst
        elif syst.startswith('bts'):
            [bts_index] = re.match(r'bts(\d+)', syst).groups()
            return int(bts_index)
        else:
            log.error(f'Invalid systematic: {syst}')
            raise
    #---------------------------------------------------------------
    def _get_signal(self, preffix, nent=None):
        log.info(f'Getting signal PDF for {preffix}')
        year, trig = self._preffix_to_info(preffix)
        log.warning(f'Using 2018 signal for {year}')
        year = '2018'

        syst       = self._get_systematic(trig, 'sig')
        sig, d_cns = hqm_model.get_signal_shape(dataset=year, trigger=trig, parameter_name_prefix=preffix, systematic=syst)

        self._add_constraints(d_cns)

        nsg    = zfit.Parameter(f'nsg_{preffix}', nent, 0, 100000)
        esig   = sig.create_extended(nsg, name=extset.sig_name)

        if   preffix.startswith('ee_'):
            self._nsig_ee = nsg 
        elif preffix.startswith('mm_'):
            self._nsig_mm = nsg 
        else:
            log.error(f'Invalid preffix: {preffix}')
            raise ValueError

        self._obs = esig.space

        return esig
    #---------------------------------------------------------------
    def _get_rare_kind(self, kind, year, trig):
        log.info(f'Getting rare PDF for {kind}/{year}/{trig}')

        syst       = self._get_systematic(trig, 'rpr')
        syst       = 0 if syst == 'nom' else syst
        if   kind == 'bpks':
            pdf = hqm_model.get_Bu2Ksee_shape(dataset= year, trigger=trig, bts_index=syst)
        elif kind == 'bdks':
            pdf = hqm_model.get_Bd2Ksee_shape(dataset= year, trigger=trig, bts_index=syst)
        elif kind == 'bsph':
            pdf = hqm_model.get_Bs2phiee_shape(dataset=year, trigger=trig, bts_index=syst)
        elif kind == 'bpk1':
            pdf = hqm_model.get_Bu2K1ee_shape(dataset= year, trigger=trig, bts_index=syst)
        elif kind == 'bpk2':
            pdf = hqm_model.get_Bu2K2ee_shape(dataset= year, trigger=trig, bts_index=syst)
        else:
            log.error(f'Invalid kind: {kind}')
            raise ValueError

        return pdf
    #---------------------------------------------------------------
    def _get_rare(self, preffix, nent=None, kind=None):
        year, trig = self._preffix_to_info(preffix)

        if year == 'all':
            d_pdf = { dset : self._get_rare_kind(kind, dset, trig) for dset in self._l_dset }
            pdf   = self._merge_year_pdf(d_pdf, preffix, kind)
        else:
            pdf   = self._get_rare_kind(kind, year, trig) 

        svl, ser = self._get_rare_scale(year, trig, kind)
        nbkg     = zfit.ComposedParameter(f'nr{kind}_{preffix}', lambda ns : ns * svl, params=[self._nsig_ee])
        pdf.set_yield(nbkg)

        self._d_rare_scl[f'{year}_{trig}_{kind}'] = [svl, ser]

        return pdf 
    #---------------------------------------------------------------
    def _merge_year_pdf(self, d_pdf, preffix, kind):
        l_dset  = list(d_pdf.keys())
        l_lumi  = [ self._d_lumi[dset] for dset in l_dset ]
        tot_lum = sum(l_lumi)
        d_frac  = { dset : lumi / tot_lum for dset, lumi in zip(l_dset, l_lumi) }

        l_lfr_par = []
        for dset in d_pdf:
            frac= d_frac[dset]
            lfr = zfit.Parameter(f'lfr_{dset}_{preffix}_{kind}', frac , 0., 1.)
            lfr.floating = False
            l_lfr_par.append(lfr)

        l_lfr_par = l_lfr_par[:-1]
        l_pdf     = [ pdf for pdf in d_pdf.values()]
        pdf       = zfit.pdf.SumPDF(l_pdf, fracs=l_lfr_par, name=kind)

        return pdf
    #---------------------------------------------------------------
    def _get_prec_dset(self, dset=None, trig=None, pref=None):
        syst = self._get_systematic(trig, 'cpr')

        if   isinstance(syst, str):
            pdf, d_cns = hqm_model.get_part_reco(dataset=dset, trigger=trig, parameter_name_prefix=pref, systematic=syst)
        elif isinstance(syst, int):
            pdf, d_cns = hqm_model.get_part_reco(dataset=dset, trigger=trig, parameter_name_prefix=pref, bts_index =syst)
        else:
            log.error(f'Invalid systematic: {syst}')
            raise

        self._add_constraints(d_cns)

        return pdf
    #---------------------------------------------------------------
    def _get_prec(self, preffix):
        log.info(f'Getting cc PRec: {preffix}')

        year, trig = self._preffix_to_info(preffix)

        if year == 'all':
            d_pdf = { dset : self._get_prec_dset(dset=dset, trig=trig, pref=preffix) for dset in self._l_dset }
            pdf   = self._merge_year_pdf(d_pdf, preffix, 'prc')
        else:
            pdf = self._get_prec_dset(dataset=year, trig=trig, pref=preffix)

        nent= self._get_entries(nsig=None, kind='prc', trigger=trig, year=year)
        npr = zfit.Parameter(f'npr_{preffix}', nent, 0, 100000)
        pdf.set_yield(npr)

        return pdf 
    #---------------------------------------------------------------
    def _get_msid(self, preffix):
        log.info(f'Getting mis-ID for: {preffix}')
        dset, trig    = self._preffix_to_info(preffix)

        dset = 'all_int' if dset == 'all' else dset

        bld= msid(name='Mis-ID', version=self._msid_vers, obs=self._obs, preffix=preffix)
        bld.load_model(dset, trig)
        pdf=bld.build_model() 

        return pdf
    #---------------------------------------------------------------
    def _get_scales(self, sig_pdf):
        s_par = sig_pdf.get_params()
        try:
            [scl] = [ par for par in s_par if 'dmu' in par.name ]
            [res] = [ par for par in s_par if 'ssg' in par.name ]
        except:
            log.error(f'Cannot extract resolution and scale from:')
            log.info(s_par)
            raise

        return scl, res
    #---------------------------------------------------------------
    def _get_pdf(self, preffix='', nent=None):
        preffix = f'{preffix}_{self._preffix}'
        log.info(f'Getting real PDF for: {preffix}')

        year, trig   = self._preffix_to_info(preffix)

        #bm = B mass, for electron fits only.
        #sp = Double swaps, added observable and PDFs to muon fits

        #bm PDFs are extended, sp ones are not
        d_pdf            = dict() 
        d_pdf['esig_bm'] = self._get_signal(preffix, nent=nent)
        d_pdf['ecmb_bm'] = self._get_combinatorial(preffix, fix_norm=False)
        if   preffix.startswith('ee_'):
            d_pdf['erbp_bm']  = self._get_rare(preffix, nent=nent, kind='bpks')
            d_pdf['erbd_bm']  = self._get_rare(preffix, nent=nent, kind='bdks')
            d_pdf['erbs_bm']  = self._get_rare(preffix, nent=nent, kind='bsph')
            #d_pdf['erb1_bm']  = self._get_rare(preffix, nent=nent, kind='bpk1')
            #d_pdf['erb2_bm']  = self._get_rare(preffix, nent=nent, kind='bpk2')
            d_pdf['eprc_bm']  = self._get_prec(preffix)
            d_pdf['emis_bm']  = self._get_msid(preffix)
        elif preffix.startswith('mm_'):
            osg_sp = dswpb(self._obs, self._obs_mm_sp, proc='sign', name='sign_sp')
            ocm_sp = dswpb(self._obs, self._obs_mm_sp, proc='cmb' , name= 'cmb_sp')

            oct_sp                = dswpb(self._obs, self._obs_mm_sp, proc='ctrl', name='ctrl_sp')
            sig                   = d_pdf['esig_bm']
            scl, res              = self._get_scales(sig)

            oct_sp.mass_scale     = scl
            oct_sp.mass_resolution= res 

            pdf_sig_bm = d_pdf['esig_bm']
            pdf_cmb_bm = d_pdf['ecmb_bm']

            pdf_ctr_bm, pdf_ctr_sp = oct_sp.get_pdf(extended=True)
            _         , pdf_sig_sp = osg_sp.get_pdf(extended=True)
            _         , pdf_cmb_sp = ocm_sp.get_pdf(extended=True)

            self._set_dswp_yield(pdf_ctr_sp, year)

            d_pdf['ectr_bm'] = pdf_ctr_bm
            d_pdf['esig_sp'] = pdf_sig_sp
            d_pdf['ectr_sp'] = pdf_ctr_sp
        else:
            log.error(f'Preffix does not start with mm_ or ee_: {preffix}')
            raise

        pdf = self._build_pdf(d_pdf, preffix)

        return pdf
    #---------------------------------------------------------------
    def _set_dswp_yield(self, pdf, year):
        nent = self._get_entries(nsig=None, kind='dswp', trigger='MTOS', year=year)
        ndsp = pdf.get_yield()
        ndsp.set_value(nent)
    #---------------------------------------------------------------
    def _build_pdf(self, d_pdf, preffix):
        if   preffix.startswith('mm_'):
            pdf = self._build_pdf_mm(d_pdf)
        elif preffix.startswith('ee_'):
            pdf = self._build_pdf_ee(d_pdf)
        else:
            log.error(f'Invalid preffix: {preffix}')
            raise

        return pdf 
    #---------------------------------------------------------------
    def _send_cmb_to_start(self, d_normal):
        val = d_normal['ecmb']

        del(d_normal['ecmb'])

        d_tmp = {'ecmb' : val}

        d_tmp.update(d_normal)

        return d_tmp
    #---------------------------------------------------------------
    def _order_pdfs(self, d_pdf):
        pdf_sig = d_pdf['esig_bm']
        pdf_cmb = d_pdf['ecmb_bm']
        pdf_ctr = d_pdf['ectr_bm']

        d_pdf_ord = dict()
        d_pdf_ord['ecmb_bm'] = pdf_cmb
        d_pdf_ord['ectr_bm'] = pdf_ctr
        d_pdf_ord['esig_bm'] = pdf_sig

        return d_pdf_ord
    #---------------------------------------------------------------
    def _build_pdf_mm(self, d_pdf):
        d_pdf_bm = { key : pdf for key, pdf in d_pdf.items() if key.endswith('_bm') }
        d_pdf_sp = { key : pdf for key, pdf in d_pdf.items() if key.endswith('_sp') }

        d_pdf_bm = self._order_pdfs(d_pdf_bm)

        l_pdf_bm = [ pdf for key, pdf in d_pdf_bm.items() ]
        l_pdf_sp = [ pdf for key, pdf in d_pdf_sp.items() ]

        pdf_bm   = zfit.pdf.SumPDF(l_pdf_bm, name='bms_pdf')
        pdf_sp   = zfit.pdf.SumPDF(l_pdf_sp, name='swp_pdf')

        return [pdf_bm, pdf_sp]
    #---------------------------------------------------------------
    def _build_pdf_ee(self, d_pdf):
        l_pdf=[]
        l_pdf.append(d_pdf['ecmb_bm'])
        l_pdf.append(d_pdf['eprc_bm'])

        l_pdf.append(d_pdf['erbp_bm'])
        l_pdf.append(d_pdf['erbd_bm'])
        l_pdf.append(d_pdf['erbs_bm'])
        #l_pdf.append(d_pdf['erb1_bm'])
        #l_pdf.append(d_pdf['erb2_bm'])
        l_pdf.append(d_pdf['emis_bm'])
        l_pdf.append(d_pdf['esig_bm'])

        pdf = zfit.pdf.SumPDF(l_pdf)

        return pdf
    #---------------------------------------------------------------
    def _get_pdf_names(self):
        d_leg = {}
        d_leg['prc' ] = r'$c\bar{c}_{prc}+\psi(2S)K^+$'
        d_leg['bpks'] = r'$B^+\to K^{*+}e^+e^-$'
        d_leg['bdks'] = r'$B^0\to K^{*0}e^+e^-$'
        d_leg['bsph'] = r'$B_s\to \phi e^+e^-$'
        d_leg['bpk1'] = r'$B^+\to K_{1}e^+e^-$'
        d_leg['bpk2'] = r'$B^+\to K_{2}e^+e^-$'

        return d_leg
    #---------------------------------------------------------------
    def _get_plt_set(self, ext_txt=None):
        d_plt_set               = {}
        d_plt_set['nbins']      = self._nbin
        d_plt_set['d_leg']      = self._get_pdf_names()
        d_plt_set['ext_text']   = ext_txt 
        d_plt_set['stacked']    = True
        d_plt_set['skip_pulls'] = False
        d_plt_set.update(self._d_plt_set)

        if 'ymax'  in self._d_plt_set:
            [mm_max, ee_max] = self._d_plt_set['ymax']
            ymax = mm_max if ext_txt.endswith('_mm') else ee_max
            d_plt_set['ymax'] = ymax

        return d_plt_set
    #---------------------------------------------------------------
    def _plot_model(self, key, mod):
        log.warning(f'Wont plot model, too time consumming')
        return
        if (self._out_dir is None) or (mod is None):
            return

        #In the case of the muon
        #There are two models
        #one for the B mass and other for
        #the dimuon mass
        log.info(f'Plotting: {key}')
        if isinstance(mod, list):
            l_mod = mod
            l_nam = ['bm', 'mm']
        else:
            l_mod = [mod]
            l_nam = ['bm']

        for pdf, nam in zip(l_mod, l_nam): 
            sam = pdf.create_sampler()
            self._plot_1d_model(key, pdf, sam, nam)
    #---------------------------------------------------------------
    def _plot_1d_model(self, key, mod, dat, nam):
        obj= zfp(data=dat, model=mod)
        d_plt_set = self._get_plt_set(ext_txt=key)
        obj.plot(**d_plt_set)

        log.info(f'Saving to: {self._mod_dir}/{key}_{nam}.png')
        plt.savefig(f'{self._mod_dir}/{key}_{nam}.png')
        plt.close('all')
    #---------------------------------------------------------------
    def _print_model(self, key, mod):
        if (self._mod_dir is None) or (mod is None):
            return

        d_const = {}
        for name in self._d_val:
            val = self._d_val[name]
            var = self._d_var[name]
            d_const[name] = [val, var]

        log.info(f'Saving to: {self._mod_dir}/{key}.txt')

        if isinstance(mod, list):
            l_mod = mod
            l_nam = ['bm', 'qm']
        else:
            l_mod = [mod]
            l_nam = ['bm']

        for pdf, nam in zip(l_mod, l_nam):
            zut.print_pdf(pdf, d_const=d_const, txt_path=f'{self._mod_dir}/{key}_{nam}.txt')
    #---------------------------------------------------------------
    def get_cons(self):
        '''
        Will return constraints on model parameters 

        Returns
        -----------
        d_val, d_var: Tuple of dictionaries pairing parameter name with value (mu of Gaussian) and 
        variance respectively.
        '''
        self._initialize()

        log.debug('-' * 20)
        log.debug(f'{"Name":<60}{"Value":<15}{"Variance":<15}')
        log.debug('-' * 20)
        for name in self._d_val:
            val = self._d_val[name]
            var = self._d_var[name]

            log.debug(f'{name:<60}{val:<15.3f}{var:<15.3f}')
        log.debug('-' * 20)

        return self._d_val, self._d_var
#---------------------------------------------------------------




import os
import re
import glob
import ROOT
import tqdm
import math
import numpy
import pprint
import pandas            as pnd
import utils_noroot      as utnr
import matplotlib.pyplot as plt

from importlib.resources import files
from scipy.interpolate   import griddata as spy_grid
from log_store           import log_store

log = log_store.add_logger(name='rk_extractor:scale_reader')
#------------------------------------------------------------
class scale_reader:
    '''
    Class used to read efficiencies under different BDT working points and
    return scale factor
    '''
    #--------------------------
    def __init__(self, wp=None, version=None, dset=None, trig=None):
        '''
        Parameters
        --------------------------
        wp (dict): Stores working point ends for the scan, i.e. {'BDT_cmb' : (0.5, 1.0)} 
        version (str): Version of efficiency file

        dset (str): Dataset, e.g. r1
        trig (str): Trigger, e.g. MTOS
        '''
        self._d_wp        = wp 
        self._dset        = dset
        self._trig        = trig
        self._vers        = version
        self._old_bdt     = 'BDT_cmb > 0.831497 && BDT_prc > 0.480751'

        self._d_lumi_wgt  = {'2011' : 1/8.7, '2012' : 2/8.7, '2015' : 0.3/8.7, '2016' : 1.6/8.7, '2017' : 1.7/8.7, '2018' : 2.1/8.7} 
        self._df          = None

        self._cmb_nom     = None
        self._prc_nom     = None

        self._cmb_new     = self._d_wp['BDT_cmb'] 
        self._prc_new     = self._d_wp['BDT_prc'] 
        self._regex       = 'BDT_cmb > ([\.,\d]+) && BDT_prc > ([\.,\d]+)'

        self._cmb_rng     = None
        self._prc_rng     = None

        self._initialized = False
    #--------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._check_wp()
        self._read_nominal_wp()

        self._initialized = True 
    #--------------------------
    def _get_min_max(self):
        min_cmb = self._df.BDT_cmb.min()
        max_cmb = self._df.BDT_cmb.max()

        min_prc = self._df.BDT_prc.min()
        max_prc = self._df.BDT_prc.max()

        self._cmb_rng = min_cmb, max_cmb
        self._prc_rng = min_prc, max_prc
    #--------------------------
    def _check_wp(self):
        try:
            low, hig = self._cmb_new
            assert hig > low
        except:
            log.error(f'Expected tuple of two ordered elements, got: {self._cmb_new}')
            raise

        try:
            low, hig = self._prc_new
            assert hig > low
        except:
            log.error(f'Expected tuple of two ordered elements, got: {self._prc_new}')
            raise
    #--------------------------
    def _read_nominal_wp(self):
        log.warning(f'Normalizing with respect of hardcoded old selection: {self._old_bdt}')
        mch = re.match(self._regex, self._old_bdt)
        if not mch:
            log.error(f'Cannot find WP in: {selection}')
            raise

        [cmb_str, prc_str] = mch.groups()

        self._cmb_nom = float(cmb_str), 10 
        self._prc_nom = float(prc_str), 10 

        log.debug(f'Using nominal WP: {self._cmb_nom}, {self._prc_nom}')
    #--------------------------
    def _load_data(self):
        if   self._dset == 'r1':
            l_year = ['2011', '2012']
        elif self._dset == 'r2p1':
            l_year = ['2015', '2016']
        elif self._dset in ['2017', '2018']:
            l_year = [self._dset]
        elif self._dset == 'all':
            l_year = ['2011', '2012', '2015', '2016', '2017', '2018']
        else:
            log.error(f'Invalid dataset: {self._dset}')
            raise

        l_df     = [ self._load_year(year) for year in l_year ]

        self._df = pnd.concat(l_df, axis=0)
    #--------------------------
    def _load_year(self, year):
        file_path = files('extractor_data').joinpath(f'sig_wgt/{self._vers}/{self._trig}_{year}.json')
        if not os.path.isfile(file_path):
            log.error(f'File not found: {file_path}')
            raise

        df        = pnd.read_json(file_path)
        df['wgt'] = df['wgt'] * self._d_lumi_wgt[year]

        return df
    #--------------------------
    def _check_range(self, t_wpt, t_bound, kind):
        if (t_wpt[0] - t_bound[0]) < -0.01:
            log.warning(f'For WP of kind {kind}, {t_wpt[0]} < {t_bound[0]}')
    #--------------------------
    def _get_key(self, t_val, name):
        val_1, val_2 = t_val

        key = f'{name}_{val_1:011.8f}_{val_2:011.8f}'

        return key
    #--------------------------
    def _read_cached_yield(self, t_cmb=None, t_prc=None):
        file_path = files('extractor_data').joinpath(f'sig_wgt/{self._vers}/yld_{self._trig}_{self._dset}.json')
        if not os.path.isfile(file_path):
            return

        cmb_key = self._get_key(t_cmb, 'BDT_cmb')
        prc_key = self._get_key(t_prc, 'BDT_prc')
        tot_key = f'{cmb_key}_{prc_key}'

        d_yield = utnr.load_json(file_path)
        if tot_key not in d_yield: 
            return

        return d_yield[tot_key]
    #--------------------------
    def _cache_yield(self, t_cmb=None, t_prc=None, value=None):
        cmb_key = self._get_key(t_cmb, 'BDT_cmb')
        prc_key = self._get_key(t_prc, 'BDT_prc')
        tot_key = f'{cmb_key}_{prc_key}'

        d_val          = {}
        d_val[tot_key] = value

        file_path = files('extractor_data').joinpath(f'sig_wgt/{self._vers}/yld_{self._trig}_{self._dset}.json')
        log.debug(f'Caching to: {file_path}')
        if os.path.isfile(file_path):
            d_val_cache = utnr.load_json(file_path)
            d_val_cache.update(d_val)
            utnr.dump_json(d_val_cache, file_path)
        else:
            utnr.dump_json(d_val      , file_path)
    #--------------------------
    def _read_yield(self, t_cmb=None, t_prc=None):
        val = self._read_cached_yield(t_cmb, t_prc)
        if val is not None:
            return val

        log.debug(f'Cached value not found, calculating from weights')

        self._load_data()
        self._get_min_max()
        self._check_range(t_cmb, self._cmb_rng, 'BDT_cmb')
        self._check_range(t_prc, self._prc_rng, 'BDT_prc')

        cmb_lo, cmb_hi = t_cmb
        prc_lo, prc_hi = t_prc

        df = self._df
        value_init = df.wgt.sum()

        cmb_cut = f'(BDT_cmb > {cmb_lo:.3f}) & (BDT_cmb < {cmb_hi:.3f})'
        prc_cut = f'(BDT_prc > {prc_lo:.3f}) & (BDT_prc < {prc_hi:.3f})'

        df = df.query(cmb_cut)
        df = df.query(prc_cut)

        value_final = df.wgt.sum()

        log.debug(f'BDT cmb: {cmb_cut}')
        log.debug(f'BDT prc: {prc_cut}')
        log.debug(f'Yield: {value_init:.0f} -> {value_final:.0f}')

        self._cache_yield(t_cmb=t_cmb, t_prc=t_prc, value=value_final)

        return value_final
    #--------------------------
    def get_scale(self):
        '''
        Returns
        -------------------
        scale (float): Yield ratio between WP and nominal.
        '''
        self._initialize()

        yld_nom = self._read_yield(t_cmb=self._cmb_nom, t_prc=self._prc_nom)
        yld_new = self._read_yield(t_cmb=self._cmb_new, t_prc=self._prc_new)
        scale   = yld_new / yld_nom
        log.debug(f'{scale:.3f}= {yld_new:.0f}/{yld_nom:.0f}')

        if scale == 0:
            new_scale = 1e-3
            log.warning(f'Found a scale of {scale}, using {new_scale}')
            return new_scale 

        return scale
#------------------------------------------------------------




import ROOT
import zfit

from rkex_model import model
from mc_reader  import mc_reader as mc_rdr
from np_reader  import np_reader as np_rdr
from cs_reader  import cs_reader as cs_rdr

import rk.utilities as rkut
import pytest
import pprint
import os

#-----------------------------
def delete_all_pars():
    d_par = zfit.Parameter._existing_params
    l_key = list(d_par.keys())

    for key in l_key:
        del(d_par[key])
#----------------------------------------------------
def rename_keys(d_data, use_txs=True):
    d_rename = {}
    if use_txs:
        d_rename[  'r1_TOS'] = d_data['d1']
        d_rename[  'r1_TIS'] = d_data['d1']

        d_rename['r2p1_TOS'] = d_data['d2']
        d_rename['r2p1_TIS'] = d_data['d2']

        d_rename['2017_TOS'] = d_data['d3']
        d_rename['2017_TIS'] = d_data['d3']

        d_rename['2018_TOS'] = d_data['d4']
        d_rename['2018_TIS'] = d_data['d4']
    else:
        d_rename[  'r1']     = d_data['d1']
        d_rename['r2p1']     = d_data['d2']
        d_rename['2017']     = d_data['d3']
        d_rename['2018']     = d_data['d4']

    return d_rename
#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#----------------------
def test_simple():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_mcmu= {'d1' : (5000, 4900), 'd2' : (5100, 4900), 'd3' : (5100, 4800), 'd4' : (5200, 5100)}
    d_mcsg= {'d1' :      (2,  4), 'd2' :     (1, 1.8), 'd3' :       (2, 3), 'd4' :       (3, 4)}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)
    d_mcmu=rename_keys(d_mcmu)
    d_mcsg=rename_keys(d_mcsg)

    mod         = model(preffix='simple', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_nent)
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_print():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_mcmu= {'d1' : (5000, 4900), 'd2' : (5100, 4900), 'd3' : (5100, 4800), 'd4' : (5200, 5100)}
    d_mcsg= {'d1' :      (2,  4), 'd2' :     (1, 1.8), 'd3' :       (2, 3), 'd4' :       (3, 4)}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)
    d_mcmu=rename_keys(d_mcmu)
    d_mcsg=rename_keys(d_mcsg)

    mod         = model(preffix='simple', d_eff=d_eff, d_nent=d_nent, l_dset=['all_TOS'])
    d_mod       = mod.get_model()
    model.show(d_mod)

    delete_all_pars()
#----------------------
def test_data():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_mcmu= {'d1' : (5000, 4900), 'd2' : (5100, 4900), 'd3' : (5100, 4800), 'd4' : (5200, 5100)}
    d_mcsg= {'d1' :      (2,  4), 'd2' :     (1, 1.8), 'd3' :       (2, 3), 'd4' :       (3, 4)}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)
    d_mcmu=rename_keys(d_mcmu)
    d_mcsg=rename_keys(d_mcsg)

    mod         = model(preffix='simple', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_nent, d_dtmu=d_mcmu, d_dtsg=d_mcsg)
    mod.out_dir = 'tests/rkex_model/simple' 
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()

    delete_all_pars()
#----------------------
def test_real():
    skip_test()

    rdr           = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache     = True 
    rdr.cache_dir = 'tests/np_reader/cache'
    d_eff         = rdr.get_eff()
    d_byld        = rdr.get_byields()
    d_byld_avg    = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld    = rkut.reso_to_rare(d_byld_avg, kind='jpsi')

    rdr           = mc_rdr(version='v4')
    rdr.cache     = False 
    d_mcmu        = rdr.get_parameter(name='mu')
    d_mcsg        = rdr.get_parameter(name='sg')

    mod           = model(preffix='real', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_nent=d_rare_yld)
    mod.out_dir   = 'tests/rkex_model/real' 
    d_dat         = mod.get_data()
    d_mod         = mod.get_model()

    delete_all_pars()
#----------------------
def test_pars():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_mcmu= {'d1' : (5280, 5270), 'd2' : (5280, 5275), 'd3' : (5278, 5282), 'd4' : (5270, 5260)}
    d_mcsg= {'d1' :    (20,  40), 'd2' :     (10, 18), 'd3' :     (20, 30), 'd4' :     (30, 40)}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)
    d_mcmu=rename_keys(d_mcmu)
    d_mcsg=rename_keys(d_mcsg)
    d_dtmu=d_mcmu
    d_dtsg=d_mcsg

    mod         = model(preffix='pars', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_dtmu=d_dtmu, d_dtsg=d_dtsg, d_nent=d_nent)
    mod.out_dir = 'tests/rkex_model/pars' 
    d_dat       = mod.get_data()
    d_mod       = mod.get_model()
    d_par       = mod.get_prefit_pars()

    pprint.pprint(d_par)

    delete_all_pars()
#----------------------
def test_const():
    d_eff = {'d1' :   (0.5, 0.4), 'd2' :   (0.4, 0.3), 'd3' :   (0.3, 0.2), 'd4' :   (0.2, 0.1)}
    d_nent= {'d1' :          1e4, 'd2' :          1e4, 'd3' :          1e4, 'd4' :          1e4}
    d_mcmu= {'d1' : (5280, 5270), 'd2' : (5280, 5275), 'd3' : (5278, 5282), 'd4' : (5270, 5260)}
    d_mcsg= {'d1' :    (20,  40), 'd2' :     (10, 18), 'd3' :     (20, 30), 'd4' :     (30, 40)}

    d_eff =rename_keys(d_eff)
    d_nent=rename_keys(d_nent, use_txs=False)
    d_mcmu=rename_keys(d_mcmu)
    d_mcsg=rename_keys(d_mcsg)
    d_dtmu=d_mcmu
    d_dtsg=d_mcsg

    rdr         = cs_rdr(version='v4', preffix='const')
    _, d_var    = rdr.get_constraints()

    rdr         = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache   = True
    cv_sys      = rdr.get_cov(kind='sys')
    cv_sta      = rdr.get_cov(kind='sta')

    mod         = model(preffix='const', d_eff=d_eff, d_mcmu=d_mcmu, d_mcsg=d_mcsg, d_dtmu=d_dtmu, d_dtsg=d_dtsg, d_nent=d_nent)
    d_par       = mod.get_prefit_pars(d_var=d_var, ck_cov=cv_sys + cv_sta)

    pprint.pprint(d_par)

    delete_all_pars()
#----------------------
def main():
    test_print()
    return
    test_simple()
    test_const()
    test_data()
    test_real()
    test_pars()
#----------------------
if __name__ == '__main__':
    main()




from setuptools import setup, find_packages

import os
import glob

#----------------------------------------------
def is_code(file_path):
    try:
        with open(file_path) as ifile:
            ifile.read()
    except:
        return False

    return True
#----------------------------------------------
def get_scripts(dir_path):
    l_obj = glob.glob(f'{dir_path}/*')
    l_scr = [ obj for obj in l_obj if is_code(obj)]

    return l_scr
#----------------------------------------------
def get_packages():
    l_pkg = find_packages(where='src') + ['']

    return l_pkg
#----------------------------------------------
def get_data_packages(pkg):
    l_pkg= [] 
    if pkg == 'extractor_data':
        l_pkg.append('config/*.toml')
        l_pkg.append('npr_data/*/*.json')
        l_pkg.append('rare_sf/*/*.json')
        l_pkg.append('sb_fits/*/*.json')
        l_pkg.append('sig_wgt/*/yld_*.json')
    else:
        raise

    return l_pkg
#----------------------------------------------
setup(
        name              = 'rk_extractor',
        version           = '1.0.6',
        description       = 'Used to extract RK from simultaneous fits',
        scripts           = get_scripts('scripts/jobs') + get_scripts('scripts/offline'),
        long_description  = '',
        packages          = get_packages(),
        package_dir       = {'' : 'src'},
        package_data      = {'extractor_data' : get_data_packages('extractor_data')}, 
        install_requires  = open('requirements.txt').read().splitlines()
        )




import os
import math
import numpy
import pprint
import jacobi            as jac
import matplotlib.pyplot as plt

from logzero import logger as log
#----------------------------------------
class combiner:
    '''
    Class used to combine CK values from multiple datasets into
    a combined one
    '''
    #----------------------------------------
    def __init__(self, rk=None, eff=None, yld=None):
        '''
        rk (float): Value of RK for which the toy test is made
        eff (dict): Holds the muon and electron full efficiencies, e.g. {'r1_TOS' : (eff_mu, eff_ee)}
        yld (dict): Holds the yields of rare decays of B mesons, before any selection, e.g. {'r1_TOS' : yld}
        '''

        self._rk     = rk 
        self._d_eff  = eff
        self._d_yld  = yld 
        self._l_dset = ['r1', 'r2p1', '2017', '2018']
        self._l_trig = ['TOS', 'TIS']

        self._out_dir     = None
        self._initialized = False
    #----------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._pad_yields()

        right_sizes = len(self._d_yld)       == len(self._d_eff) 
        right_keys  =     self._d_yld.keys() ==     self._d_eff.keys()

        if not right_sizes:
            log.error(f'Sizes of inputs is wrong:')
            pprint.pprint(self._d_yld)
            pprint.pprint(self._d_eff)
            raise

        if not right_keys:
            log.error(f'Sizes of input dictionaries are different')
            raise

        self._initialized = True
    #----------------------------------------
    def _pad_yields(self):
        d_yld = {}
        for key, val in self._d_yld.items():
            d_yld[f'{key}_TOS'] = val
            d_yld[f'{key}_TIS'] = val

        self._d_yld = d_yld
    #----------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot make directory: {value}')
            raise
        self._out_dir = value
    #----------------------------------------
    def _get_ck(self):
        l_ck = []

        for trig in self._l_trig:
            tot_ee, tot_mm = 0, 0
            for dset in self._l_dset:
                key = f'{dset}_{trig}'
                eff_mm, eff_ee = self._d_eff[key]
                yld            = self._d_yld[key]

                tot_mm += eff_mm * yld
                tot_ee += eff_ee * yld

            ck   = self._rk * tot_ee / tot_mm
            l_ck.append(ck)

        return l_ck 
    #----------------------------------------
    def get_combination(self):
        '''
        Parameters
        ------------
        add_tis (bool): If true, will return TIS ck and 2x2 covariance, if false, TIS is dropped

        Returns
        ------------
        Tuple with dictionaries with the rjpsi and efficiencies, e.g:

        d_rjpsi, d_eff, cov = cmb.get_cobination()

        the values of the first two dictionaries themselves are not meaningful, but when used
        to calculate ck, they should provide the right combined value
        '''
        self._initialize()

        [ck_val_tos, ck_val_tis] = self._get_ck()

        eff_val = 1 - 1e-6
        ck_var  = 1e-4

        d_rjpsi = {'all_TOS' :                    1., 'all_TIS' :                    1. }
        d_eff   = {'all_TOS' : (eff_val, ck_val_tos), 'all_TIS' : (eff_val, ck_val_tis) }
        cov     = [[ck_var, 0], [0, ck_var]]

        return d_rjpsi, d_eff, numpy.array(cov)
#----------------------------------------

